"""
Getting the data
================

In this section, we will dicuss how to get functional data to
use in scikit-fda. We will briefly describe the
:class:`~skfda.representation.grid.FDataGrid` class, which is the type that
scikit-fda uses for storing and working with functional data in discretized
form. We will discuss also how to import functional data from several sources
and show how to fetch and load existing datasets popular in the :term:`FDA`
literature.

.. Disable isort
    isort:skip_file

"""

# Author: Carlos Ramos Carreño
# License: MIT
#
# sphinx_gallery_thumbnail_number = 6

##############################################################################
# The FDataGrid class
# -------------------
#
# In order to use scikit-fda, first we need functional data to analyze.
# A common case is to have each functional observation measured at the same
# points.
# This kind of functional data is easily representable in scikit-fda using
# the :class:`~skfda.representation.grid.FDataGrid` class.
# The :class:`~skfda.representation.grid.FDataGrid` has two important
# attributes: ``data_matrix`` and ``grid_points``.
#
# The attribute ``grid_points`` is a tuple with the same length as the
# number of domain dimensions (that is, one for curves, two for surfaces...).
# Each of its elements is a 1D numpy :class:`~numpy.ndarray` containing the
# grid points for that particular dimension,
#  .. math::
#     ((t_1, \ldots, t_{M_i}))_{i=1}^p,
# where :math:`M_i` is the number of measurement points for each "argument"
# or domain coordinate of the function :math:`i` and :math:`p` is the domain
# dimension.
#
# The attribute ``data_matrix`` is a
# numpy :class:`~numpy.ndarray` containing the measured values of the
# functions in the grid spanned by the grid points. For functions
# :math:`\{x_i: \mathbb{R}^p \to \mathbb{R}^q\}_{i=1}^N` this is a tensor
# with dimensions :math:`N \times M_1 \times \ldots \times M_p \times q`.

##############################################################################
# In order to create a :class:`~skfda.representation.grid.FDataGrid`, these
# attributes may be provided. The attributes are converted to
# :class:`~numpy.ndarray` when necessary.
#
# .. note::
#
#     The grid points can be omitted,
#     and in that case their number is inferred from the dimensions of
#     ``data_matrix`` and they are automatically assigned as equispaced points
#     in the unitary cube in the domain set.
#
#     In the common case of functions with domain dimension of 1, the list of
#     grid points can be passed directly as ``grid_points``.
#
#     If the codomain dimension is 1, the last dimension of ``data_matrix``
#     can be dropped.

##############################################################################
# The following example shows the creation of a
# :class:`~skfda.representation.grid.FDataGrid` with two functions (curves)
# :math:`\{x_i: \mathbb{R} \to \mathbb{R}\}, i=1,2` measured at the same
# (non-equispaced) points.

import skfda
import matplotlib.pyplot as plt

grid_points = [0, 0.2, 0.5, 0.9, 1]  # Grid points of the curves
data_matrix = [
    [0, 0.2, 0.5, 0.9, 1],     # First observation
    [0, 0.04, 0.25, 0.81, 1],  # Second observation
]

fd = skfda.FDataGrid(
    data_matrix=data_matrix,
    grid_points=grid_points,
)

fd.plot()
plt.show()

##############################################################################
# Advanced example
# ^^^^^^^^^^^^^^^^
#
# In order to better understand the FDataGrid structure, you can consider the
# following example, in which a :class:`~skfda.representation.grid.FDataGrid`
# object is created, containing just one function (vector-valued surface)
# :math:`x: \mathbb{R}^2 \to \mathbb{R}^4`.


grid_points_surface = [
    [0.2, 0.5, 0.7],      # Measurement points in first domain dimension
    [0, 1.5],             # Measurement points in second domain dimension
]

data_matrix_surface = [
    # First observation
    [
        # 0.2
        [
            # Value at (0.2, 0)
            [1, 2, 3, 4.1],
            # Value at (0.2, 1.5)
            [0, 1, -1.3, 2],
        ],
        # 0.5
        [
            # Value at (0.5, 0)
            [-2, 0, 5.5, 7],
            # Value at (0.5, 1.5)
            [2, 1.1, -1, -2],
        ],
        # 0.7
        [
            # Value at (0.7, 0)
            [0, 0, 1.1, 1],
            # Value at (0.7, 1.5)
            [-3, 5, -0.5, -2],
        ],
    ],
    # This example has only one observation. Next observations would be
    # added here.
]

fd = skfda.FDataGrid(
    data_matrix=data_matrix_surface,
    grid_points=grid_points_surface,
)

fd.plot()
plt.show()

##############################################################################
# Importing data
# --------------
#
# Usually one does not construct manually the functions, but instead uses
# measurements already formatted in a common format, such as comma-separated
# values (CSV), attribute-relation file format (ARFF) or Matlab and R formats.
#
# If your data is in one of these formats, you can import it into a numpy
# array using the IO functions available in
# `Numpy <https://numpy.org/devdocs/reference/routines.io.html>`_ (for simple
# text-based or binary formats, such as CSV) or in
# `Scipy <https://docs.scipy.org/doc/scipy/reference/io.html>`_ (for Matlab,
# Fortran or ARFF files). For importing data in the R format one can also
# use the package `RData <https://rdata.readthedocs.io>`_ with is already a
# dependency of scikit-fda, as it is used to load the example datasets.

##############################################################################
# Once your data has been introduced as a :class:`~numpy.ndarray` instance,
# you will need to give it the proper dimensions and use it to instantiate
# a functional data object.

##############################################################################
# .. note::
#
#     :class:`Pandas DataFrames <pandas.DataFrame>` are also popular as
#     datasets containers in the Python scientific ecosystem. If you have
#     data in a Pandas DataFrame, you can extract its content as a Numpy
#     array using the method :meth:`~pandas.DataFrame.to_numpy` of the
#     DataFrame.

##############################################################################
# As an example, we will load the
# :func:`digits dataset <sklearn.datasets.load_digits>` of scikit-learn, which
# is a preprocessed subset of the MNIST dataset, containing digit images. The
# data is already a numpy array. As the data has been flattened into a 1D
# vector of pixels, we need to reshape the arrays to their original 8x8 shape.
# Then this array can be used to construct the digits as surfaces.

from sklearn.datasets import load_digits

X, y = load_digits(return_X_y=True)
X = X.reshape(-1, 8, 8)

fd = skfda.FDataGrid(X)

# Plot the first 2 observations
fd[0].plot()
fd[1].plot()
plt.show()


##############################################################################
# Common datasets
# ---------------
#
# scikit-fda can download and import for you several of the most popular
# datasets in the :term:`FDA` literature, such as the Berkeley Growth
# dataset (function :func:`~skfda.datasets.fetch_growth`) or the Canadian
# Weather dataset (function :func:`~skfda.datasets.fetch_weather`). These
# datasets are often useful as benchmarks, in order to compare results
# between different algorithms, or simply as examples to use in teaching or
# research.

X, y = skfda.datasets.fetch_growth(return_X_y=True)

X.plot(group=y)
plt.show()

##############################################################################
# Datasets from CRAN
# ^^^^^^^^^^^^^^^^^^
#
# If you want to work with a dataset for which no fetching function exist, and
# you know that is available inside a R package in the CRAN repository, you
# can try using the function :func:`~skfda.datasets.fetch_cran`. This function
# will load the package, fetch the dataset and convert it to Python objects
# using the packages
# `scikit-datasets <https://github.com/daviddiazvico/scikit-datasets>`_ and
# `RData <https://rdata.readthedocs.io>`_. As datasets in CRAN follow no
# particular structure, you will need to know how it is structured internally
# in order to use it properly.

##############################################################################
# .. note::
#
#     Functional data objects from some packages, such as
#     `fda.usc <https://cran.r-project.org/web/packages/fda.usc/index.html>`_
#     are automatically recognized as such and converted to
#     :class:`~skfda.representation.grid.FDataGrid` instances. This
#     behaviour can be disabled or customized to work with more packages.

data = skfda.datasets.fetch_cran("MCO", "fda.usc")

data["MCO"]["intact"].plot()
plt.show()

##############################################################################
# Datasets from the UEA & UCR Time Series Classification Repository
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# The `UEA & UCR Time Series Classification Repository
# <http://www.timeseriesclassification.com/>`_ is a popular repository
# for classification problems involving time series data. The datasets used
# can be considered also as functional observations, where the functions
# involved have domain dimension of 1, and the grid points are
# equispaced. Thus, they have also been used in the :term:`FDA` literature.
# The original UCR datasets are univariate time series, while the new UEA
# datasets incorporate also vector-valued data.
# In scikit-fda, the function :func:`~skfda.datasets.fetch_ucr` can be used
# to obtain both kinds of datasets as
# :class:`~skfda.representation.grid.FDataGrid` instances.

# Load ArrowHead dataset from UCR
dataset = skfda.datasets.fetch_ucr("ArrowHead")
dataset["data"].plot()
plt.show()

##############################################################################

# Load BasicMotions dataset from UEA
dataset = skfda.datasets.fetch_ucr("BasicMotions")
dataset["data"].plot()
plt.show()

##############################################################################
# Synthetic data
# --------------
#
# Sometimes it is not enough to have real-world data at your disposal.
# Perhaps the messy nature of real-world data makes difficult to detect when
# a particular algorithm has a strange behaviour. Perhaps you want to see how
# it performs under a simplified model. Maybe you want to see what happens
# when your data has particular characteristics, for which no dataset is
# available. Or maybe you only want to illustrate a concept without having
# to introduce a particular set of data.
#
# In those cases, the ability to use generated data is desirable. To aid this
# use case, scikit-learn provides several functions that generate data
# according to some model. These functions are in the
# :doc:`datasets </modules/datasets>` module and have the prefix ``make_``.
# Maybe the most useful of those are the functions
# :func:`skfda.datasets.make_gaussian_process` and
# :func:`skfda.datasets.make_gaussian` which can be used to generate Gaussian
# processes and Gaussian fields with different covariance functions.

import numpy as np

cov = skfda.misc.covariances.Exponential(length_scale=0.1)

fd = skfda.datasets.make_gaussian_process(
    start=0,
    stop=4,
    n_samples=5,
    n_features=100,
    mean=lambda t: np.power(t, 2),
    cov=cov,
)

fd.plot()
plt.show()

##############################################################################
# In order to know all the available functionalities to load existing and
# synthetic datasets it is recommended to look at the documentation of the
# :doc:`datasets </modules/datasets>` module.



"""Functional Principal Component Analysis Module."""

from __future__ import annotations

from typing import Callable, Optional, TypeVar, Union

import numpy as np
import scipy.integrate
from scipy.linalg import solve_triangular
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.decomposition import PCA

from ....misc import inner_product_matrix
from ....misc.metrics import l2_norm
from ....misc.regularization import L2Regularization, compute_penalty_matrix
from ....representation import FData
from ....representation._typing import ArrayLike
from ....representation.basis import Basis, FDataBasis
from ....representation.grid import FDataGrid

Function = TypeVar("Function", bound=FData)
WeightsCallable = Callable[[np.ndarray], np.ndarray]


class FPCA(
    BaseEstimator,  # type: ignore
    TransformerMixin,  # type: ignore
):
    r"""
    Principal component analysis.

    Class that implements functional principal component analysis for both
    basis and grid representations of the data. Most parameters are shared
    when fitting a FDataBasis or FDataGrid, except weights and
    ``components_basis``.

    Parameters:
        n_components: Number of principal components to keep from
            functional principal component analysis. Defaults to 3.
        centering: If ``True`` then calculate the mean of the functional
            data object and center the data first. Defaults to ``True``.
        regularization: Regularization object to be applied.
        components_basis: The basis in which we want the principal
            components. We can use a different basis than the basis contained
            in the passed FDataBasis object. This parameter is only used when
            fitting a FDataBasis.
        weights: the weights vector used for
            discrete integration. If none then the trapezoidal rule is used for
            computing the weights. If a callable object is passed, then the
            weight vector will be obtained by evaluating the object at the
            sample points of the passed FDataGrid object in the fit method.
            This parameter is only used when fitting a FDataGrid.

    Attributes:
        components\_ (FData): this contains the principal components in a
            basis representation.
        explained_variance\_ (array_like): The amount of variance explained by
            each of the selected components.
        explained_variance_ratio\_ (array_like): this contains the percentage
            of variance explained by each principal component.
        singular_values\_: The singular values corresponding to each of the
            selected components. 
        mean\_ (FData): mean of the train data.


    Examples:
        Construct an artificial FDataBasis object and run FPCA with this
        object. The resulting principal components are not compared because
        there are several equivalent possibilities.

        >>> import skfda
        >>> data_matrix = np.array([[1.0, 0.0], [0.0, 2.0]])
        >>> grid_points = [0, 1]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>> basis = skfda.representation.basis.Monomial(
        ...     domain_range=(0,1), n_basis=2
        ... )
        >>> basis_fd = fd.to_basis(basis)
        >>> fpca_basis = FPCA(2)
        >>> fpca_basis = fpca_basis.fit(basis_fd)

        In this example we apply discretized functional PCA with some simple
        data to illustrate the usage of this class. We initialize the
        FPCA object, fit the artificial data and obtain the scores.
        The results are not tested because there are several equivalent
        possibilities.

        >>> data_matrix = np.array([[1.0, 0.0], [0.0, 2.0]])
        >>> grid_points = [0, 1]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>> fpca_grid = FPCA(2)
        >>> fpca_grid = fpca_grid.fit(fd)

    """

    def __init__(
        self,
        n_components: int = 3,
        centering: bool = True,
        regularization: Optional[L2Regularization[FData]] = None,
        weights: Optional[Union[ArrayLike, WeightsCallable]] = None,
        components_basis: Optional[Basis] = None,
    ) -> None:
        self.n_components = n_components
        self.centering = centering
        self.regularization = regularization
        self.weights = weights
        self.components_basis = components_basis

    def _center_if_necessary(
        self,
        X: Function,
        *,
        learn_mean: bool = True,
    ) -> Function:

        if learn_mean:
            self.mean_ = X.mean()

        return X - self.mean_ if self.centering else X

    def _fit_basis(
        self,
        X: FDataBasis,
        y: None = None,
    ) -> FPCA:
        """
        Compute the first n_components principal components and saves them.

        The eigenvalues associated with these principal components are also
        saved. For more details about how it is implemented please view the
        referenced book.

        Args:
            X: The functional data object to be analysed.
            y: Ignored.

        Returns:
            self

        References:
            .. [RS05-8-4-2] Ramsay, J., Silverman, B. W. (2005). Basis function
                expansion of the functions. In *Functional Data Analysis*
                (pp. 161-164). Springer.

        """
        # the maximum number of components is established by the target basis
        # if the target basis is available.
        n_basis = (
            self.components_basis.n_basis
            if self.components_basis
            else X.basis.n_basis
        )
        n_samples = X.n_samples
        # necessary in inverse_transform
        self.n_samples_ = X.n_samples

        # check that the number of components is smaller than the sample size
        if self.n_components > X.n_samples:
            raise AttributeError(
                "The sample size must be bigger than the "
                "number of components",
            )

        # check that we do not exceed limits for n_components as it should
        # be smaller than the number of attributes of the basis
        if self.n_components > n_basis:
            raise AttributeError(
                "The number of components should be "
                "smaller than the number of attributes of "
                "target principal components' basis.",
            )

        # if centering is True then subtract the mean function to each function
        # in FDataBasis
        X = self._center_if_necessary(X)

        # setup principal component basis if not given
        components_basis = self.components_basis
        if components_basis is not None:
            # First fix domain range if not already done
            components_basis = components_basis.copy(
                domain_range=X.basis.domain_range,
            )
            g_matrix = components_basis.gram_matrix()
            # The matrix that are in charge of changing the computed principal
            # components to target matrix is essentially the inner product
            # of both basis.
            j_matrix = X.basis.inner_product_matrix(components_basis)
        else:
            # If no other basis is specified we use the same basis as the
            # passed FDataBasis object
            components_basis = X.basis.copy()
            g_matrix = components_basis.gram_matrix()
            j_matrix = g_matrix

        self._X_basis = X.basis
        self._j_matrix = j_matrix

        # Apply regularization / penalty if applicable
        regularization_matrix = compute_penalty_matrix(
            basis_iterable=(components_basis,),
            regularization_parameter=1,
            regularization=self.regularization,
        )

        # apply regularization
        if regularization_matrix is not None:
            g_matrix = (g_matrix + regularization_matrix)

        # obtain triangulation using cholesky
        l_matrix = np.linalg.cholesky(g_matrix)

        # we need L^{-1} for a multiplication, there are two possible ways:
        # using solve to get the multiplication result directly or just invert
        # the matrix. We choose solve because it is faster and more stable.
        # The following matrix is needed: L^{-1}*J^T
        l_inv_j_t = solve_triangular(
            l_matrix,
            np.transpose(j_matrix),
            lower=True,
        )

        # the final matrix, C(L-1Jt)t for svd or (L-1Jt)-1CtC(L-1Jt)t for PCA
        final_matrix = (
            X.coefficients @ np.transpose(l_inv_j_t)
        )

        # initialize the pca module provided by scikit-learn
        pca = PCA(n_components=self.n_components)
        pca.fit(final_matrix)

        # we choose solve to obtain the component coefficients for the
        # same reason: it is faster and more efficient
        component_coefficients = solve_triangular(
            np.transpose(l_matrix),
            np.transpose(pca.components_),
            lower=False,
        )

        self.explained_variance_ratio_ = pca.explained_variance_ratio_
        self.explained_variance_ = pca.explained_variance_
        self.singular_values_ = pca.singular_values_
        self.components_ = X.copy(
            basis=components_basis,
            coefficients=component_coefficients.T,
            sample_names=(None,) * self.n_components,
        )

        return self

    def _transform_basis(
        self,
        X: FDataBasis,
        y: None = None,
    ) -> np.ndarray:
        """Compute the n_components first principal components score.

        Args:
            X: The functional data object to be analysed.
            y: Ignored.

        Returns:
            Principal component scores.

        """
        if X.basis != self._X_basis:
            raise ValueError(
                "The basis used in fit is different from "
                "the basis used in transform.",
            )

        # in this case it is the inner product of our data with the components
        return (
            X.coefficients @ self._j_matrix
            @ self.components_.coefficients.T
        )

    def _fit_grid(
        self,
        X: FDataGrid,
        y: None = None,
    ) -> FPCA:
        r"""
        Compute the n_components first principal components and saves them.

        The eigenvalues associated with these principal
        components are also saved. For more details about how it is implemented
        please view the referenced book, chapter 8.

        In summary, we are performing standard multivariate PCA over
        :math:`\frac{1}{\sqrt{N}} \mathbf{X} \mathbf{W}^{1/2}` where :math:`N`
        is the number of samples in the dataset, :math:`\mathbf{X}` is the data
        matrix and :math:`\mathbf{W}` is the weight matrix (this matrix
        defines the numerical integration). By default the weight matrix is
        obtained using the trapezoidal rule.

        Args:
            X: The functional data object to be analysed.
            y: Ignored.

        Returns:
            self.

        References:
            .. [RS05-8-4-1] Ramsay, J., Silverman, B. W. (2005). Discretizing
                the functions. In *Functional Data Analysis* (p. 161).
                Springer.

        """
        # check that the number of components is smaller than the sample size
        if self.n_components > X.n_samples:
            raise AttributeError(
                "The sample size must be bigger than the "
                "number of components",
            )

        # check that we do not exceed limits for n_components as it should
        # be smaller than the number of attributes of the funcional data object
        if self.n_components > X.data_matrix.shape[1]:
            raise AttributeError(
                "The number of components should be "
                "smaller than the number of discretization "
                "points of the functional data object.",
            )

        # data matrix initialization
        fd_data = X.data_matrix.reshape(X.data_matrix.shape[:-1])

        # get the number of samples and the number of points of descretization
        n_samples, n_points_discretization = fd_data.shape

        # necessary for inverse_transform
        self.n_samples_ = n_samples

        # if centering is True then subtract the mean function to each function
        # in FDataBasis
        X = self._center_if_necessary(X)

        # establish weights for each point of discretization
        if not self.weights:
            # grid_points is a list with one array in the 1D case
            # in trapezoidal rule, suppose \deltax_k = x_k - x_{k-1}, the
            # weight vector is as follows: [\deltax_1/2, \deltax_1/2 +
            # \deltax_2/2, \deltax_2/2 + \deltax_3/2, ... , \deltax_n/2]
            identity = np.eye(len(X.grid_points[0]))
            self.weights = scipy.integrate.simps(identity, X.grid_points[0])
        elif callable(self.weights):
            self.weights = self.weights(X.grid_points[0])
            # if its a FDataGrid then we need to reduce the dimension to 1-D
            # array
            if isinstance(self.weights, FDataGrid):
                self.weights = np.squeeze(self.weights.data_matrix)

        weights_matrix = np.diag(self.weights)

        basis = FDataGrid(
            data_matrix=np.identity(n_points_discretization),
            grid_points=X.grid_points,
        )

        regularization_matrix = compute_penalty_matrix(
            basis_iterable=(basis,),
            regularization_parameter=1,
            regularization=self.regularization,
        )

        basis_matrix = basis.data_matrix[..., 0]
        if regularization_matrix is not None:
            basis_matrix = basis_matrix + regularization_matrix

        fd_data = np.linalg.solve(
            basis_matrix.T,
            fd_data.T,
        ).T

        # see docstring for more information
        final_matrix = fd_data @ np.sqrt(weights_matrix)

        pca = PCA(n_components=self.n_components)
        pca.fit(final_matrix)
        self.components_ = X.copy(
            data_matrix=np.transpose(
                np.linalg.solve(
                    np.sqrt(weights_matrix),
                    np.transpose(pca.components_),
                ),
            ),
            sample_names=(None,) * self.n_components,
        )

        self.explained_variance_ratio_ = pca.explained_variance_ratio_
        self.explained_variance_ = pca.explained_variance_
        self.singular_values_ = pca.singular_values_

        return self

    def _transform_grid(
        self,
        X: FDataGrid,
        y: None = None,
    ) -> np.ndarray:
        """
        Compute the ``n_components`` first principal components score.

        Args:
            X: The functional data object to be analysed.
            y: Ignored.

        Returns:
            Principal component scores.

        """
        # in this case its the coefficient matrix multiplied by the principal
        # components as column vectors

        return (
            X.data_matrix.reshape(X.data_matrix.shape[:-1])
            * self.weights
            @ np.transpose(
                self.components_.data_matrix.reshape(
                    self.components_.data_matrix.shape[:-1],
                ),
            )
        )

    def fit(
        self,
        X: FData,
        y: None = None,
    ) -> FPCA:
        """
        Compute the n_components first principal components and saves them.

        Args:
            X: The functional data object to be analysed.
            y: Ignored.

        Returns:
            self

        """
        if isinstance(X, FDataGrid):
            return self._fit_grid(X, y)
        elif isinstance(X, FDataBasis):
            return self._fit_basis(X, y)

        raise AttributeError("X must be either FDataGrid or FDataBasis")

    def transform(
        self,
        X: FData,
        y: None = None,
    ) -> np.ndarray:
        """
        Compute the ``n_components`` first principal components scores.

        Args:
            X: The functional data object to be analysed.
            y: Only present because of fit function convention

        Returns:
            Principal component scores.

        """
        X = self._center_if_necessary(X, learn_mean=False)

        if isinstance(X, FDataGrid):
            return self._transform_grid(X, y)
        elif isinstance(X, FDataBasis):
            return self._transform_basis(X, y)

        raise AttributeError("X must be either FDataGrid or FDataBasis")

    def fit_transform(
        self,
        X: FData,
        y: None = None,
    ) -> np.ndarray:
        """
        Compute the n_components first principal components and their scores.

        Args:
            X: The functional data object to be analysed.
            y: Ignored

        Returns:
            Principal component scores.

        """
        return self.fit(X, y).transform(X, y)

    def inverse_transform(
        self,
        pc_scores: np.ndarray,
    ) -> FData:
        """
        Compute the recovery from the fitted principal components scores.

        In other words,
        it maps ``pc_scores``, from the fitted functional PCs' space,
        back to the input functional space.
        ``pc_scores`` might be an array returned by ``transform`` method.

        Args:
            pc_scores: ndarray (n_samples, n_components).

        Returns:
            A FData object.

        """
        # check the instance is fitted.

        # input format check:
        if isinstance(pc_scores, np.ndarray):
            if pc_scores.ndim == 1:
                pc_scores = pc_scores[np.newaxis, :]

            if pc_scores.shape[1] != self.n_components:
                raise AttributeError(
                    "pc_scores must be a numpy array "
                    "with n_samples rows and n_components columns.",
                )
        else:
            raise AttributeError("pc_scores is not a numpy array.")

        # inverse_transform is slightly different whether
        # .fit was applied to FDataGrid or FDataBasis object
        # Does not work (boundary problem in x_hat and bias reconstruction)
        if isinstance(self.components_, FDataGrid):

            additional_args = {
                "data_matrix": np.einsum(
                    'nc,c...->n...',
                    pc_scores,
                    self.components_.data_matrix,
                ),
            }

        elif isinstance(self.components_, FDataBasis):

            additional_args = {
                "coefficients": pc_scores @ self.components_.coefficients,
            }

        return self.mean_.copy(
            **additional_args,
            sample_names=(None,) * len(pc_scores),
        ) + self.mean_



from __future__ import annotations

from typing import Any, Callable, Generator, NoReturn, Union

import numpy as np
import pandas
import pytest
from pandas import Series
from pandas.api.extensions import ExtensionArray, ExtensionDtype
from pandas.tests.extension import base

import skfda
from skfda.representation.grid import FDataGrid


##############################################################################
# Fixtures
##############################################################################
@pytest.fixture
def dtype() -> ExtensionDtype:
    """Return the ExtensionDtype to validate."""
    return skfda.representation.grid.FDataGridDType(
        grid_points=[
            np.arange(10),
            np.arange(10) / 10,
        ],
        dim_codomain=3,
    )


@pytest.fixture
def data() -> ExtensionArray:
    """
    Return data.

    Length-100 array for this type.
    * data[0] and data[1] should both be non missing
    * data[0] and data[1] should not be equal

    """
    data_matrix = np.arange(1, 100 * 10 * 10 * 3 + 1).reshape(100, 10, 10, 3)
    grid_points = [
        np.arange(10),
        np.arange(10) / 10,
    ]

    return skfda.FDataGrid(data_matrix, grid_points=grid_points)


@pytest.fixture
def data_for_twos() -> ExtensionArray:
    """Return a length-100 array in which all the elements are two."""
    data_matrix = np.full(
        100 * 10 * 10 * 3, fill_value=2,
    ).reshape(100, 10, 10, 3)
    grid_points = [
        np.arange(10),
        np.arange(10) / 10,
    ]

    return skfda.FDataGrid(data_matrix, grid_points=grid_points)


@pytest.fixture
def data_missing() -> ExtensionArray:
    """Return a length-2 array with [NA, Valid]."""
    data_matrix = np.arange(
        2 * 10 * 10 * 3,
        dtype=np.float_,
    ).reshape(2, 10, 10, 3)
    data_matrix[0, ...] = np.NaN
    grid_points = [
        np.arange(10),
        np.arange(10) / 10,
    ]

    return skfda.FDataGrid(data_matrix, grid_points=grid_points)


@pytest.fixture(params=["data", "data_missing"])
def all_data(
    request,
    data: ExtensionArray,
    data_missing: ExtensionArray,
) -> ExtensionArray:
    """Return 'data' or 'data_missing'."""
    if request.param == "data":
        return data
    elif request.param == "data_missing":
        return data_missing


@pytest.fixture
def data_repeated(
    data: ExtensionArray,
) -> Callable[[int], Generator[ExtensionArray, None, None]]:
    """
    Generate many datasets.

    Args:
        data : Fixture implementing `data`

    Returns:
        Callable[[int], Generator]:
            A callable that takes a `count` argument and
            returns a generator yielding `count` datasets.
    """

    def gen(count: int) -> Generator[ExtensionArray, None, None]:
        yield from (
            data for _ in range(count)
        )

    return gen


@pytest.fixture
def data_for_sorting() -> NoReturn:
    """
    Return ength-3 array with a known sort order.

    This should be three items [B, C, A] with
    A < B < C
    """
    raise NotImplementedError


@pytest.fixture
def data_missing_for_sorting() -> NoReturn:
    """
    Return length-3 array with a known sort order.
    This should be three items [B, NA, A] with
    A < B and NA missing.
    """
    raise NotImplementedError


@pytest.fixture
def na_cmp() -> Callable[..., bool]:
    """
    Binary operator for comparing NA values.

    Should return a function of two arguments that returns
    True if both arguments are (scalar) NA for your type.
    By default, uses ``operator.is_``
    """
    def isna(
        x: Union[pandas.NA, FDataGrid],
        y: Union[pandas.NA, FDataGrid],
    ) -> bool:
        return (
            (x is pandas.NA or all(x.isna()))
            and (y is pandas.NA or all(y.isna()))
        )

    return isna


@pytest.fixture
def na_value() -> pandas.NA:
    """Return the scalar missing value for this type. Default 'None'."""
    return pandas.NA


@pytest.fixture
def data_for_grouping() -> NoReturn:
    """
    Return data for factorization, grouping, and unique tests.

    Expected to be like [B, B, NA, NA, A, A, B, C]
    Where A < B < C and NA is missing
    """
    raise NotImplementedError


@pytest.fixture(params=[True, False])
def box_in_series(request) -> bool:
    """Whether to box the data in a Series."""
    return request.param


@pytest.fixture(
    params=[
        lambda x: 1,
        lambda x: [1] * len(x),
        lambda x: Series([1] * len(x)),
        lambda x: x,
    ],
    ids=["scalar", "list", "series", "object"],
)
def groupby_apply_op(request) -> Callable[[FDataGrid], Any]:
    """Functions to test groupby.apply()."""
    return request.param


@pytest.fixture(params=[True, False])
def as_frame(request) -> bool:
    """Whether to support Series and Series.to_frame() comparison testing."""
    return request.param


@pytest.fixture(params=[True, False])
def as_series(request) -> bool:
    """Boolean fixture to support arr and Series(arr) comparison testing."""
    return request.param


@pytest.fixture(params=[True, False])
def use_numpy(request) -> bool:
    """
    Compare ExtensionDtype and numpy.

    Boolean fixture to support comparison testing of ExtensionDtype array
    and numpy array.
    """
    return request.param


@pytest.fixture(params=["ffill", "bfill"])
def fillna_method(request) -> str:
    """
    Series.fillna parameter fixture.

    Parametrized fixture giving method parameters 'ffill' and 'bfill' for
    Series.fillna(method=<method>) testing.
    """
    return request.param


@pytest.fixture(params=[True, False])
def as_array(request) -> bool:
    """Whether to support ExtensionDtype _from_sequence method testing."""
    return request.param


_all_arithmetic_operators = [
    "__add__",
    "__radd__",
    "__sub__",
    "__rsub__",
    "__mul__",
    "__rmul__",
    #     "__floordiv__",
    #     "__rfloordiv__",
    "__truediv__",
    "__rtruediv__",
    #     "__pow__",
    #     "__rpow__",
    #     "__mod__",
    #     "__rmod__",
]


@pytest.fixture(params=_all_arithmetic_operators)
def all_arithmetic_operators(request) -> Callable[..., Any]:
    """
    Fixture for dunder names for common arithmetic operations.
    """
    return request.param


@pytest.fixture(params=["__eq__", "__ne__",
                        # "__le__", "__lt__", "__ge__", "__gt__"
                        ])
def all_compare_operators(request) -> Callable[..., Any]:
    """
    Fixture for dunder names for common compare operations
    """
    return request.param


_all_numeric_reductions = [
    "sum",
    #     "max",
    #     "min",
    "mean",
    #     "prod",
    #     "std",
    #     "var",
    #     "median",
    #     "kurt",
    #     "skew",
]


@pytest.fixture(params=_all_numeric_reductions)
def all_numeric_reductions(request):
    """
    Fixture for numeric reduction names.
    """
    return request.param

##############################################################################
# Tests
##############################################################################


class TestCasting(base.BaseCastingTests):

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_astype_str(self) -> None:
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_astype_string(self) -> None:
        pass


class TestConstructors(base.BaseConstructorsTests):

    # Does not support scalars which are also ExtensionArrays
    @pytest.mark.skip(reason="Unsupported")
    def test_series_constructor_scalar_with_index(self) -> None:
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_from_dtype(self) -> None:
        pass


class TestDtype(base.BaseDtypeTests):

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_construct_from_string_own_name(self) -> None:
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_is_dtype_from_name(self) -> None:
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_eq_with_str(self) -> None:
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_construct_from_string(
        self,
        dtype: ExtensionDtype,
    ) -> None:
        pass


class TestGetitem(base.BaseGetitemTests):
    pass


class TestInterface(base.BaseInterfaceTests):

    # Does not support scalars which are also array_like
    @pytest.mark.skip(reason="Unsupported")
    def test_array_interface(self) -> None:
        pass

    # We do not implement setitem
    @pytest.mark.skip(reason="Unsupported")
    def test_copy(
        self,
        dtype: ExtensionDtype,
    ) -> None:
        pass

    # We do not implement setitem
    @pytest.mark.skip(reason="Unsupported")
    def test_view(
        self,
        dtype: ExtensionDtype,
    ) -> None:
        pass

    # Pending https://github.com/pandas-dev/pandas/issues/38812 resolution
    @pytest.mark.skip(reason="Bugged")
    def test_contains(
        self,
        data: ExtensionArray,
        data_missing: ExtensionArray,
    ) -> None:
        pass


class TestArithmeticOps(base.BaseArithmeticOpsTests):

    series_scalar_exc = None

    # Bug introduced by https://github.com/pandas-dev/pandas/pull/37132
    @pytest.mark.skip(reason="Unsupported")
    def test_arith_frame_with_scalar(
        self,
        data: ExtensionArray,
        all_arithmetic_operators: Callable[..., Any],
    ) -> None:
        pass

    # Does not convert properly a list of FData to a FData
    @pytest.mark.skip(reason="Unsupported")
    def test_arith_series_with_array(
        self,
        dtype: ExtensionDtype,
    ) -> None:
        pass

    # Does not error on operations
    @pytest.mark.skip(reason="Unsupported")
    def test_error(
        self,
        dtype: ExtensionDtype,
    ) -> None:
        pass


class TestComparisonOps(base.BaseComparisonOpsTests):

    # Cannot be compared with 0
    @pytest.mark.skip(reason="Unsupported")
    def test_compare_scalar(
        self,
        data: ExtensionArray,
        all_compare_operators: Callable[..., Any],
    ) -> None:
        pass

    # Not sure how to pass it. Should it be reimplemented?
    @pytest.mark.skip(reason="Unsupported")
    def test_compare_array(
        self,
        data: ExtensionArray,
        all_compare_operators: Callable[..., Any],
    ) -> None:
        pass


class TestNumericReduce(base.BaseNumericReduceTests):

    def check_reduce(
        self,
        s: FDataGrid,
        op_name: str,
        skipna: bool,
    ) -> None:
        result = getattr(s, op_name)(skipna=skipna)
        assert result.n_samples == 1



"""
Depth Measures Module.

This module includes different methods to order functional data,
from the center (larger values) outwards(smaller ones).

"""
from __future__ import annotations

import itertools
from typing import Optional

import numpy as np
import scipy.integrate

from ... import FDataGrid
from .multivariate import Depth, SimplicialDepth, _UnivariateFraimanMuniz


class IntegratedDepth(Depth[FDataGrid]):
    r"""
    Functional depth as the integral of a multivariate depth.

    Args:
        multivariate_depth (Depth): Multivariate depth to integrate.
            By default it is the one used by Fraiman and Muniz, that is,

            .. math::
                D(x) = 1 - \left\lvert \frac{1}{2}- F(x)\right\rvert

    Examples:
        >>> import skfda
        >>>
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> depth = skfda.exploratory.depth.IntegratedDepth()
        >>> depth(fd)
        array([ 0.5  ,  0.75 ,  0.925,  0.875])

    References:
        Fraiman, R., & Muniz, G. (2001). Trimmed means for functional
        data. Test, 10(2), 419–440. https://doi.org/10.1007/BF02595706


    """

    def __init__(
        self,
        *,
        multivariate_depth: Optional[Depth[np.ndarray]] = None,
    ) -> None:
        self.multivariate_depth = multivariate_depth

    def fit(  # noqa: D102
        self,
        X: FDataGrid,
        y: None = None,
    ) -> IntegratedDepth:

        self.multivariate_depth_: Depth[np.ndarray]

        if self.multivariate_depth is None:
            self.multivariate_depth_ = _UnivariateFraimanMuniz()
        else:
            self.multivariate_depth_ = self.multivariate_depth

        self._domain_range = X.domain_range
        self._grid_points = X.grid_points
        self.multivariate_depth_.fit(X.data_matrix)
        return self

    def transform(self, X: FDataGrid) -> np.ndarray:  # noqa: D102

        pointwise_depth = self.multivariate_depth_.transform(X.data_matrix)

        interval_len = (
            self._domain_range[0][1]
            - self._domain_range[0][0]
        )

        integrand = pointwise_depth

        for d, s in zip(X.domain_range, X.grid_points):
            integrand = scipy.integrate.simps(
                integrand,
                x=s,
                axis=1,
            )
            interval_len = d[1] - d[0]
            integrand /= interval_len

        return integrand

    @property  # noqa: WPS125
    def max(self) -> float:  # noqa: WPS125
        if self.multivariate_depth is None:
            return 1

        return self.multivariate_depth.max

    @property  # noqa: WPS125
    def min(self) -> float:  # noqa: WPS125
        if self.multivariate_depth is None:
            return 1 / 2

        return self.multivariate_depth.min


class ModifiedBandDepth(IntegratedDepth):
    """
    Implementation of Modified Band Depth for functional data.

    The band depth of each sample is obtained by computing the fraction of time
    its graph is contained in the bands determined by two sample curves.
    In the case the fdatagrid :term:`domain` dimension is 2, instead of curves,
    surfaces determine the bands. In larger dimensions, the hyperplanes
    determine the bands.

    Examples:
        >>> import skfda
        >>>
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> depth = skfda.exploratory.depth.ModifiedBandDepth()
        >>> values = depth(fd)
        >>> values.round(2)
        array([ 0.5 ,  0.83,  0.73,  0.67])

    References:
        López-Pintado, S., & Romo, J. (2009). On the Concept of
        Depth for Functional Data. Journal of the American Statistical
        Association, 104(486), 718–734.
        https://doi.org/10.1198/jasa.2009.0108
    """

    def __init__(self) -> None:
        super().__init__(multivariate_depth=SimplicialDepth())


class BandDepth(Depth[FDataGrid]):
    """
    Implementation of Band Depth for functional data.

    The band depth of each sample is obtained by computing the fraction of the
    bands determined by two sample curves containing the whole graph of the
    first one. In the case the fdatagrid :term:`domain` dimension is 2, instead
    of curves, surfaces determine the bands. In larger dimensions, the
    hyperplanes determine the bands.

    Examples:
        >>> import skfda
        >>>
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> depth = skfda.exploratory.depth.BandDepth()
        >>> depth(fd)
        array([ 0.5       ,  0.83333333,  0.5       ,  0.5       ])

    References:
        López-Pintado, S., & Romo, J. (2009). On the Concept of
        Depth for Functional Data. Journal of the American Statistical
        Association, 104(486), 718–734.
        https://doi.org/10.1198/jasa.2009.0108

    """

    def fit(self, X: FDataGrid, y: None = None) -> BandDepth:  # noqa: D102

        if X.dim_codomain != 1:
            raise NotImplementedError(
                "Band depth not implemented for vector valued functions",
            )

        self._distribution = X
        return self

    def transform(self, X: FDataGrid) -> np.ndarray:  # noqa: D102

        num_in = 0
        n_total = 0

        for f1, f2 in itertools.combinations(self._distribution, 2):
            between_range_1 = (
                (f1.data_matrix <= X.data_matrix)
                & (X.data_matrix <= f2.data_matrix)
            )

            between_range_2 = (
                (f2.data_matrix <= X.data_matrix)
                & (X.data_matrix <= f1.data_matrix)
            )

            between_range = between_range_1 | between_range_2

            num_in += np.all(
                between_range,
                axis=tuple(range(1, X.data_matrix.ndim)),
            )
            n_total += 1

        return num_in / n_total



from __future__ import annotations

from typing import Callable

import numpy as np

import scipy.integrate

from ...representation import FData
from ._operators import Operator


class IntegralTransform(Operator[FData, Callable[[np.ndarray], np.ndarray]]):
    """Integral operator.

    Parameters:
        kernel_function:  Kernel function corresponding to the operator.

    """

    def __init__(
        self,
        kernel_function: Callable[[np.ndarray, np.ndarray], np.ndarray],
    ) -> None:
        self.kernel_function = kernel_function

    def __call__(  # noqa: D102
        self,
        f: FData,
    ) -> Callable[[np.ndarray], np.ndarray]:

        def evaluate_covariance(  # noqa: WPS430
            points: np.ndarray,
        ) -> np.ndarray:

            def integral_body(  # noqa: WPS430
                integration_var: np.ndarray,
            ) -> np.ndarray:
                return (
                    f(integration_var)
                    * self.kernel_function(integration_var, points)
                )

            domain_range = f.domain_range[0]

            return scipy.integrate.quad_vec(
                integral_body,
                domain_range[0],
                domain_range[1],
            )[0]

        return evaluate_covariance



from . import covariances, kernels, lstsq, metrics, operators, regularization
from ._math import (
    cosine_similarity,
    cosine_similarity_matrix,
    cumsum,
    exp,
    inner_product,
    inner_product_matrix,
    log,
    log2,
    log10,
    sqrt,
)



"""Implementation of Lp norms."""
import math
from builtins import isinstance
from typing import Union

import numpy as np
import scipy.integrate
from typing_extensions import Final

from skfda.representation._typing import NDArrayFloat

from ...representation import FData, FDataBasis
from ...representation._typing import NDArrayFloat
from ._typing import Norm


class LpNorm():
    r"""
    Norm of all the observations in a FDataGrid object.

    For each observation f the Lp norm is defined as:

    .. math::
        \| f \| = \left( \int_D \| f \|^p dx \right)^{
        \frac{1}{p}}

    Where D is the :term:`domain` over which the functions are defined.

    The integral is approximated using Simpson's rule.

    In general, if f is a multivariate function :math:`(f_1, ..., f_d)`, and
    :math:`D \subset \mathbb{R}^n`, it is applied the following generalization
    of the Lp norm.

    .. math::
        \| f \| = \left( \int_D \| f \|_{*}^p dx \right)^{
        \frac{1}{p}}

    Where :math:`\| \cdot \|_*` denotes a vectorial norm. See
    :func:`vectorial_norm` to more information.

    For example, if :math:`f: \mathbb{R}^2 \rightarrow \mathbb{R}^2`, and
    :math:`\| \cdot \|_*` is the euclidean norm
    :math:`\| (x,y) \|_* = \sqrt{x^2 + y^2}`, the lp norm applied is

    .. math::
        \| f \| = \left( \int \int_D \left ( \sqrt{ \| f_1(x,y)
        \|^2 + \| f_2(x,y) \|^2 } \right )^p dxdy \right)^{
        \frac{1}{p}}

    The objects ``l1_norm``, ``l2_norm`` and ``linf_norm`` are instances of
    this class with commonly used values of ``p``, namely 1, 2 and infinity.

    Args:
        p: p of the lp norm. Must be greater or equal
            than 1. If ``p=math.inf`` it is used the L infinity metric.
            Defaults to 2.
        vector_norm: vector norm to apply. If it is a float, is the index of
            the multivariate lp norm. Defaults to the same as ``p``.

    Examples:
        Calculates the norm of a FDataGrid containing the functions y = 1
        and y = x defined in the interval [0,1].

        >>> import skfda
        >>> import numpy as np
        >>>
        >>> x = np.linspace(0, 1, 1001)
        >>> fd = skfda.FDataGrid([np.ones(len(x)), x] ,x)
        >>> norm = skfda.misc.metrics.LpNorm(2)
        >>> norm(fd).round(2)
        array([ 1.  ,  0.58])

        As the norm with `p=2` is a common choice, one can use `l2_norm`
        directly:

        >>> skfda.misc.metrics.l2_norm(fd).round(2)
        array([ 1.  ,  0.58])

        The lp norm is only defined if p >= 1.

        >>> norm = skfda.misc.metrics.LpNorm(0.5)
        Traceback (most recent call last):
            ....
        ValueError: p (=0.5) must be equal or greater than 1.

    """

    def __init__(
        self,
        p: float,
        vector_norm: Union[Norm[NDArrayFloat], float, None] = None,
    ) -> None:

        # Checks that the lp normed is well defined
        if not np.isinf(p) and p < 1:
            raise ValueError(f"p (={p}) must be equal or greater than 1.")

        self.p = p
        self.vector_norm = vector_norm

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"p={self.p}, vector_norm={self.vector_norm})"
        )

    def __call__(self, vector: Union[NDArrayFloat, FData]) -> NDArrayFloat:
        """Compute the Lp norm of a functional data object."""
        from ...misc import inner_product

        if isinstance(vector, np.ndarray):
            return np.linalg.norm(vector, ord=self.p, axis=-1)

        vector_norm = self.vector_norm

        if vector_norm is None:
            vector_norm = self.p

        # Special case, the inner product is heavily optimized
        if self.p == vector_norm == 2:
            return np.sqrt(inner_product(vector, vector))

        if isinstance(vector, FDataBasis):
            if self.p != 2:
                raise NotImplementedError

            start, end = vector.domain_range[0]
            integral = scipy.integrate.quad_vec(
                lambda x: np.power(np.abs(vector(x)), self.p),
                start,
                end,
            )
            res = np.sqrt(integral[0]).flatten()

        else:
            data_matrix = vector.data_matrix
            original_shape = data_matrix.shape
            data_matrix = data_matrix.reshape(-1, original_shape[-1])

            data_matrix = (np.linalg.norm(
                vector.data_matrix,
                ord=vector_norm,
                axis=-1,
                keepdims=True,
            ) if isinstance(vector_norm, (float, int))
                else vector_norm(data_matrix)
            )
            data_matrix = data_matrix.reshape(original_shape[:-1] + (1,))

            if np.isinf(self.p):

                res = np.max(
                    data_matrix,
                    axis=tuple(range(1, data_matrix.ndim)),
                )

            elif vector.dim_domain == 1:

                # Computes the norm, approximating the integral with Simpson's
                # rule.
                res = scipy.integrate.simps(
                    data_matrix[..., 0] ** self.p,
                    x=vector.grid_points,
                ) ** (1 / self.p)

            else:
                # Needed to perform surface integration
                return NotImplemented

        if len(res) == 1:
            return res[0]

        return res


l1_norm: Final = LpNorm(1)
l2_norm: Final = LpNorm(2)
linf_norm: Final = LpNorm(math.inf)


def lp_norm(
    vector: Union[NDArrayFloat, FData],
    *,
    p: float,
    vector_norm: Union[Norm[NDArrayFloat], float, None] = None,
) -> NDArrayFloat:
    r"""Calculate the norm of all the observations in a FDataGrid object.

    For each observation f the Lp norm is defined as:

    .. math::
        \| f \| = \left( \int_D \| f \|^p dx \right)^{
        \frac{1}{p}}

    Where D is the :term:`domain` over which the functions are defined.

    The integral is approximated using Simpson's rule.

    In general, if f is a multivariate function :math:`(f_1, ..., f_d)`, and
    :math:`D \subset \mathbb{R}^n`, it is applied the following generalization
    of the Lp norm.

    .. math::
        \| f \| = \left( \int_D \| f \|_{*}^p dx \right)^{
        \frac{1}{p}}

    Where :math:`\| \cdot \|_*` denotes a vectorial norm. See
    :func:`vectorial_norm` to more information.

    For example, if :math:`f: \mathbb{R}^2 \rightarrow \mathbb{R}^2`, and
    :math:`\| \cdot \|_*` is the euclidean norm
    :math:`\| (x,y) \|_* = \sqrt{x^2 + y^2}`, the lp norm applied is

    .. math::
        \| f \| = \left( \int \int_D \left ( \sqrt{ \| f_1(x,y)
        \|^2 + \| f_2(x,y) \|^2 } \right )^p dxdy \right)^{
        \frac{1}{p}}

    Note:
        This function is a wrapper of :class:`LpNorm`, available only for
        convenience. As the parameter ``p`` is mandatory, it cannot be used
        where a fully-defined norm is required: use an instance of
        :class:`LpNorm` in those cases.

    Args:
        vector: Vector object.
        p: p of the lp norm. Must be greater or equal
            than 1. If ``p=math.inf`` it is used the L infinity metric.
            Defaults to 2.
        vector_norm: vector norm to apply. If it is a float, is the index of
            the multivariate lp norm. Defaults to the same as ``p``.

    Returns:
        numpy.darray: Matrix with as many rows as observations in the first
        object and as many columns as observations in the second one. Each
        element (i, j) of the matrix is the inner product of the ith
        observation of the first object and the jth observation of the second
        one.

    Examples:
        Calculates the norm of a FDataGrid containing the functions y = 1
        and y = x defined in the interval [0,1].

        >>> import skfda
        >>> import numpy as np
        >>>
        >>> x = np.linspace(0,1,1001)
        >>> fd = skfda.FDataGrid([np.ones(len(x)), x] ,x)
        >>> skfda.misc.metrics.lp_norm(fd, p=2).round(2)
        array([ 1.  ,  0.58])

        As the norm with ``p=2`` is a common choice, one can use ``l2_norm``
        directly:

        >>> skfda.misc.metrics.l2_norm(fd).round(2)
        array([ 1.  ,  0.58])

        The lp norm is only defined if p >= 1.

        >>> skfda.misc.metrics.lp_norm(fd, p=0.5)
        Traceback (most recent call last):
            ....
        ValueError: p (=0.5) must be equal or greater than 1.

    See also:
        :class:`LpNorm`

    """
    return LpNorm(p=p, vector_norm=vector_norm)(vector)



from __future__ import annotations

from typing import Callable, Tuple

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.linear_model import LogisticRegression as mvLogisticRegression
from sklearn.utils.validation import check_is_fitted

from ..._utils import _classifier_get_classes
from ...representation import FDataGrid
from ...representation._typing import NDArrayAny, NDArrayInt


class LogisticRegression(
    BaseEstimator,  # type: ignore
    ClassifierMixin,  # type: ignore
):
    r"""Logistic Regression classifier for functional data.

    This class implements the sequential “greedy” algorithm
    for functional logistic regression proposed in
    :footcite:ts:`bueno++_2021_functional`.

    .. warning::
        For now, only binary classification for functional
        data with one dimensional domains is supported.

    Args:
        p:
            number of points (and coefficients) to be selected by
            the algorithm.

    Attributes:
        classes\_: A list containing the name of the classes
        points\_: A list containing the selected points.
        coef\_: A list containing the coefficient for each selected point.
        intercept\_: Independent term.

    Examples:
        >>> from numpy import array
        >>> from skfda.datasets import make_gaussian_process
        >>> from skfda.ml.classification import LogisticRegression
        >>> fd1 = make_gaussian_process(
        ...         n_samples=50,
        ...         n_features=100,
        ...         noise=0.7,
        ...         random_state=0,
        ... )
        >>> fd2 = make_gaussian_process(
        ...         n_samples=50,
        ...         n_features = 100,
        ...         mean = array([1]*100),
        ...         noise = 0.7,
        ...         random_state=0
        ... )
        >>> fd = fd1.concatenate(fd2)
        >>> y = 50*[0] + 50*[1]
        >>> lr = LogisticRegression()
        >>> _ = lr.fit(fd[::2], y[::2])
        >>> lr.coef_.round(2)
        array([[ 1.28,  1.17,  1.27,  1.27,  0.96]])
        >>> lr.points_.round(2)
        array([ 0.11,  0.06,  0.07,  0.03,  0.  ])
        >>> lr.score(fd[1::2],y[1::2])
        0.94

        References:
            .. footbibliography::

    """

    def __init__(
        self,
        p: int = 5,
    ) -> None:

        self.p = p

    def fit(  # noqa: D102
        self,
        X: FDataGrid,
        y: NDArrayAny,
    ) -> LogisticRegression:

        X, classes, y_ind = self._argcheck_X_y(X, y)

        self.classes_ = classes

        n_samples = len(y)
        n_features = len(X.grid_points[0])

        selected_indexes = np.zeros(self.p, dtype=np.intc)

        # multivariate logistic regression
        mvlr = mvLogisticRegression(penalty='l2')

        x_mv = np.zeros((n_samples, self.p))
        LL = np.zeros(n_features)
        for q in range(self.p):
            for t in range(n_features):

                x_mv[:, q] = X.data_matrix[:, t, 0]
                mvlr.fit(x_mv[:, :q + 1], y_ind)

                # log-likelihood function at t
                log_probs = mvlr.predict_log_proba(x_mv[:, :q + 1])
                log_probs = np.concatenate(
                    (log_probs[y_ind == 0, 0], log_probs[y_ind == 1, 1]),
                )
                LL[t] = np.mean(log_probs)

            tmax = np.argmax(LL)
            selected_indexes[q] = tmax
            x_mv[:, q] = X.data_matrix[:, tmax, 0]

        # fit for the complete set of points
        mvlr.fit(x_mv, y_ind)

        self.coef_ = mvlr.coef_
        self.intercept_ = mvlr.intercept_
        self._mvlr = mvlr

        self._selected_indexes = selected_indexes
        self.points_ = X.grid_points[0][selected_indexes]

        return self

    def predict(self, X: FDataGrid) -> NDArrayInt:  # noqa: D102
        check_is_fitted(self)
        return self._wrapper(self._mvlr.predict, X)

    def predict_log_proba(self, X: FDataGrid) -> NDArrayInt:  # noqa: D102
        check_is_fitted(self)
        return self._wrapper(self._mvlr.predict_log_proba, X)

    def predict_proba(self, X: FDataGrid) -> NDArrayInt:  # noqa: D102
        check_is_fitted(self)
        return self._wrapper(self._mvlr.predict_proba, X)

    def _argcheck_X(  # noqa: N802
        self,
        X: FDataGrid,
    ) -> FDataGrid:

        if X.dim_domain > 1:
            raise ValueError(
                f'The dimension of the domain has to be one'
                f'; got {X.dim_domain} dimensions',
            )

        return X

    def _argcheck_X_y(  # noqa: N802
        self,
        X: FDataGrid,
        y: NDArrayAny,
    ) -> Tuple[FDataGrid, NDArrayAny, NDArrayAny]:

        X = self._argcheck_X(X)

        classes, y_ind = _classifier_get_classes(y)

        if classes.size > 2:
            raise ValueError(
                f'The number of classes has to be two'
                f'; got {classes.size} classes',
            )

        if (len(y) != len(X)):
            raise ValueError(
                "The number of samples on independent variables"
                " and classes should be the same",
            )

        return (X, classes, y_ind)

    def _wrapper(
        self,
        method: Callable[[NDArrayAny], NDArrayAny],
        X: FDataGrid,
    ) -> NDArrayAny:
        """Wrap multivariate logistic regression method.

        This function transforms functional data in order to pass
        them to a multivariate logistic regression method.

        .. warning::
            This function can't be called before fit.
        """
        X = self._argcheck_X(X)

        x_mv = X.data_matrix[:, self._selected_indexes, 0]

        return method(x_mv)



# -*- coding: utf-8 -*-
"""Registration methods base class.

This module contains the abstract base class for all registration methods.

"""
from __future__ import annotations

from abc import abstractmethod
from typing import Any, TypeVar, overload

from ... import FData
from ..._utils import (
    BaseEstimator,
    InductiveTransformerMixin,
    TransformerMixin,
)

SelfType = TypeVar("SelfType")
Input = TypeVar("Input", bound=FData)
Output = TypeVar("Output", bound=FData)


class RegistrationTransformer(
    BaseEstimator,
    TransformerMixin[Input, Output, None],
):
    """Base class for the registration methods."""

    def fit(
        self: SelfType,
        X: Input,
        y: None = None,
    ) -> SelfType:
        """
        Fit the registration model.

        Args:
            X: Original (unregistered) training data.
            y: Ignored.

        Returns:
            Returns the instance itself.

        """
        return self

    @overload  # type: ignore[misc]
    def fit_transform(
        self,
        X: Input,
        y: None = None,
    ) -> Output:
        pass

    def fit_transform(
        self,
        X: Input,
        y: None = None,
        **fit_params: Any,
    ) -> Output:
        """
        Fit the registration model and return the registered data.

        Args:
            X: Original (unregistered) training data.
            y: Ignored.
            fit_params: Additional fit parameters.

        Returns:
            Registered training data.

        """
        return super().fit_transform(  # type: ignore[call-arg]
            X,
            y,
            **fit_params,
        )

    def score(self, X: Input, y: None = None) -> float:
        r"""
        Return the percentage of total variation removed.

        Computes the squared multiple correlation index of the proportion of
        the total variation due to phase, defined as:

        .. math::
            R^2 = \frac{\text{MSE}_{phase}}{\text{MSE}_{total}},

        where :math:`\text{MSE}_{total}` is the mean squared error and
        :math:`\text{MSE}_{phase}` is the mean squared error due to the phase
        explained by the registration procedure. See
        :class:`~.validation.AmplitudePhaseDecomposition` for a detailed
        explanation.

        Args:
            X: Functional data to be registered
            y: Ignored, only for API conventions.

        Returns:
            Registration score.

        See also:
            :class:`~.validation.AmplitudePhaseDecomposition`
            :class:`~.validation.LeastSquares`
            :class:`~.validation.SobolevLeastSquares`
            :class:`~.validation.PairwiseCorrelation`

        """
        from .validation import AmplitudePhaseDecomposition

        return AmplitudePhaseDecomposition()(self, X, y)


class InductiveRegistrationTransformer(
    RegistrationTransformer[Input, Output],
    InductiveTransformerMixin[Input, Output, None],
):

    @abstractmethod
    def transform(
        self: SelfType,
        X: Input,
    ) -> Output:
        """
        Register new data.

        Args:
            X: Original (unregistered) data.

        Returns:
            Registered data.

        """
        pass



"""
Introduction
============

In this section, we will briefly explain what is
:term:`functional data analysis (FDA) <FDA>`, and we will introduce
scikit-fda, a library that provides FDA tools to staticians, engineers or
machine learning practicioners in the Python scientific ecosystem.

.. Disable isort
    isort:skip_file

"""

# Author: Carlos Ramos Carreño
# License: MIT

##############################################################################
# What is functional data analysis?
# ---------------------------------
#
# Traditional multivariate statistics focus on the simultaneous analysis of
# a finite number of variables. In this setting, we have several observations,
# each of them consisting on a vector of measured values. The variables, or
# coordinates of this vector, could be correlated, but otherwise they can be
# arbitrarily ordered inside the observed vector, provided that the order is
# the same for each observation. Usually these observations are considered
# to be instances of a random vector, and a big part of the analysis is
# centered in finding the distribution associated with it.
#
# In contrast, in functional data analysis each observation is a function of
# one or several variables, such as curves or surfaces. These functions are
# usually continuous and often they are smooth, and derivatives can be
# computed. The number of variables of these objects is then infinite, as each
# evaluation of the function at one point could be considered as one variable.
# Moreover, now it is not possible to reorder the variables of the
# observations without altering substantially its structure. If the functions
# are continuous, nearby variables are highly correlated, a characteristic
# that makes some classical multivariate methods unsuitable to work with this
# data.
#
# In this setting, observations can also be considered to be instances
# of a "functional random variable", usually called a stochastic process or
# a random field. However, some of the concepts that proved very useful to
# analyze multivariate data, such as density functions, are not applicable
# to :term:`functional data`, while new tools, such as taking derivatives,
# become available.
#
# As such, functional data can benefit of a separate analysis from
# multivariate statistics, but also adapting and extending multivariate
# techniques when possible.

##############################################################################
# What is scikit-fda?
# -------------------
#
# scikit-fda is a Python library containing classes and functions that allow
# you to perform functional data analysis tasks. Using it you can:
#
#     - Represent functions as Python objects, both in a discretized fashion
#       and as a basis expansion.
#     - Apply preprocessing methods to functional data, including smoothing,
#       registration and dimensionality reduction.
#     - Perform a complete exploratory analysis of the data, summarizing its
#       main properties, detecting possible outliers and visualizing the data
#       in several ways.
#     - Apply statistical inference tools developed for functional data, such
#       as functional ANOVA.
#     - Perform usual machine learning tasks, such as classification,
#       regression or clustering, using functional observations.
#     - Combine the tools offered by scikit-fda with other tools of the Python
#       scientific ecosystem, such as those provided by the popular machine
#       learning library `scikit-learn <https://scikit-learn.org/>`_.


##############################################################################
# Anatomy of a function
# ---------------------
#
# We would like to briefly remind the reader the basic concepts that are
# employed to talk about functions. Functions in math are a relation between
# two sets, the :term:`domain` and the :term:`codomain` in which each element
# of the :term:`domain` is restricted to be related to exactly one element of
# the :term:`codomain`. The intuition behind this is that a function
# represents some type of deterministic process, that takes elements of the
# :term:`domain` as inputs and produces elements of the :term:`codomain` as
# outputs.
#
# In :term:`FDA`, the inputs or parameters of a function are assumed to be
# continuous parameters, and so are the outputs, or values of the function.
# Thus, it is usual to restrict our functional observations to be functions
# :math:`\{x_i: \mathcal{T} \subseteq \mathbb{R}^p \to \mathbb{R}^q\}_{i=1}^N`.
# In this case both the domain and codomain are (subsets of) vector spaces of
# real numbers, and one could talk of the dimension of each of them as a
# vector space (in this case the domain dimension is :math:`p` and the
# codomain dimension is :math:`q`).
#
# The most common case of functional observation, and the one that has
# received more attention in the functional data literature, is the case of
# functions
# :math:`\{x_i: \mathcal{T} \subseteq \mathbb{R} \to \mathbb{R}\}_{i=1}^N`
# (curves or trajectories).

##############################################################################
# As an example, the following code shows the Berkeley Growth dataset, one
# of the classical datasets used in :term:`FDA`. The curves are heights of
# 93 boys and girls measured at several points since their birth to
# their 18th birthday. Here the domain :math:`\mathcal{T}` is the interval
# :math:`[0, 18]` and both the domain and codomain have a dimension of one.

import skfda
import matplotlib.pyplot as plt

X, y = skfda.datasets.fetch_growth(return_X_y=True)

X.plot()
plt.show()

##############################################################################
# Functions where the domain dimension is greater than one (
# such as surfaces or higher dimensional objects) are referred to as functions
# of several variables. Functions where the codomain dimension is greater than
# one are called vector-valued functions.

##############################################################################
# As an example we show another popular dataset: Canadian Weather. Here each
# observation correspond to data taken from a different weather station in
# Canada. For each day of the year we have two values: the average temperature
# at that day among several years and the average precipitation among the same
# years. Thus, here the domain :math:`\mathcal{T}` is the interval
# :math:`[0, 365)`, the domain dimension is one and the codomain dimension
# is two. We can see that by default each coordinate of the values of the
# function is plotted as a separate coordinate function.

X, y = skfda.datasets.fetch_weather(return_X_y=True)

X.plot()
plt.show()



"""
This module contains the definition of the constants used in the package.
The following constants are defined:
.. data:: BASIS_MIN_FACTOR
    Constant used in the discretization of a basis object, by default de
    number of points used are the maximum between
    BASIS_MIN_FACTOR * n_basis + 1 and N_POINTS_FINE_MESH.
.. data:: N_POINTS_FINE_MESH
    Constant used in the discretization of a basis object, by default de
    number of points used are the maximum between
    BASIS_MIN_FACTOR * n_basis + 1 and N_POINTS_FINE_MESH.
.. data:: N_POINTS_COARSE_MESH
    Constant used in the default discretization of a basis in some methods.
.. data:: N_POINTS_UNIDIMENSIONAL_PLOT_MESH
    Number of points used in the evaluation of a function to be plotted.
.. data:: N_POINTS_SURFACE_PLOT_AX
    Number of points per axis used in the evaluation of a surface to be
    plotted.
"""

BASIS_MIN_FACTOR = 10

N_POINTS_COARSE_MESH = 201

N_POINTS_FINE_MESH = 501

N_POINTS_SURFACE_PLOT_AX = 30

N_POINTS_UNIDIMENSIONAL_PLOT_MESH = 501



from ._real_datasets import (
    fdata_constructor,
    fetch_aemet,
    fetch_cran,
    fetch_gait,
    fetch_growth,
    fetch_handwriting,
    fetch_medflies,
    fetch_octane,
    fetch_phoneme,
    fetch_tecator,
    fetch_ucr,
    fetch_weather,
)
from ._samples_generators import (
    make_gaussian,
    make_gaussian_process,
    make_multimodal_landmarks,
    make_multimodal_samples,
    make_random_warping,
    make_sinusoidal_process,
)



"""Landmark Registration of functional data module.

This module contains methods to perform the landmark registration.
"""
from __future__ import annotations

import warnings
from typing import Any, Callable, Optional, Sequence, Union

import numpy as np

from ...representation import FData, FDataGrid
from ...representation._typing import ArrayLike, GridPointsLike, NDArrayFloat
from ...representation.extrapolation import ExtrapolationLike
from ...representation.interpolation import SplineInterpolation

_FixedLocation = Union[float, Sequence[float]]
_LocationCallable = Callable[[np.ndarray], _FixedLocation]


def landmark_shift_deltas(
    fd: FData,
    landmarks: ArrayLike,
    location: Union[_FixedLocation, _LocationCallable, None] = None,
) -> NDArrayFloat:
    r"""Return the corresponding shifts to align the landmarks of the curves.

        Let :math:`t^*` the time where the landmarks of the curves will be
        aligned, and :math:`t_i` the location of the landmarks for each curve.
        The function will calculate the corresponding :math:`\delta_i` shuch
        that :math:`t_i = t^* + \delta_i`.

        This procedure will work independent of the dimension of the
        :term:`domain` and the :term:`codomain`.

    Args:
        fd: Functional data object.
        landmarks: List with the landmarks of the samples.
        location: Defines where
            the landmarks will be alligned. If a number or list is passed the
            landmarks will be alligned to it. In case of a callable is
            passed the location will be the result of the the call, the
            function should be accept as an unique parameter a numpy array
            with the list of landmarks.
            By default it will be used as location the mean of the original
            locations of the landmarks.

    Returns:
        Array containing the corresponding shifts.

    Raises:
        ValueError: If the list of landmarks does not match with the number of
            samples.

    Examples:
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import landmark_shift_deltas

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, random_state=1)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        >>> landmarks = landmarks.squeeze()

        The function will return the corresponding shifts

        >>> shifts = landmark_shift_deltas(fd, landmarks)
        >>> shifts.round(3)
        array([ 0.327, -0.173, -0.154])

        The registered samples can be obtained with a shift

        >>> fd.shift(shifts)
        FDataGrid(...)

    """
    landmarks = np.atleast_1d(landmarks)

    if len(landmarks) != fd.n_samples:
        raise ValueError(
            f"landmark list ({len(landmarks)}) must have the same"
            f" length than the number of samples ({fd.n_samples})",
        )

    loc_array: Union[float, Sequence[float], NDArrayFloat]

    # Parses location
    if location is None:
        loc_array = np.mean(landmarks)
    elif callable(location):
        loc_array = location(landmarks)
    else:
        loc_array = location

    loc_array = np.atleast_1d(loc_array)

    return landmarks - loc_array


def landmark_shift(
    *args: Any,
    **kwargs: Any,
) -> FDataGrid:

    warnings.warn(
        "Function 'landmark_shift' has been renamed. "
        "Use 'landmark_shift_registration' instead.",
        DeprecationWarning,
    )

    return landmark_shift_registration(*args, **kwargs)


def landmark_shift_registration(
    fd: FData,
    landmarks: ArrayLike,
    location: Union[_FixedLocation, _LocationCallable, None] = None,
    *,
    restrict_domain: bool = False,
    extrapolation: Optional[ExtrapolationLike] = None,
    grid_points: Optional[GridPointsLike] = None,
) -> FDataGrid:
    r"""
    Perform a shift of the curves to align the landmarks.

    Let :math:`t^*` the time where the landmarks of the curves will be
    aligned, :math:`t_i` the location of the landmarks for each curve
    and :math:`\delta_i= t_i - t^*`.

    The registered samples will have their feature aligned.

    .. math::
        x_i^*(t^*)=x_i(t^* + \delta_i)=x_i(t_i)

    Args:
        fd: Functional data object.
        landmarks: List with the landmarks of the samples.
        location: Defines where
            the landmarks will be alligned. If a numeric value is passed the
            landmarks will be alligned to it. In case of a callable is
            passed the location will be the result of the the call, the
            function should be accept as an unique parameter a numpy array
            with the list of landmarks.
            By default it will be used as location :math:`\frac{1}{2}(max(
            \text{landmarks})+ min(\text{landmarks}))` wich minimizes the
            max shift.
        restrict_domain: If True restricts the domain to
            avoid evaluate points outside the domain using extrapolation.
            Defaults uses extrapolation.
        extrapolation: Controls the
            extrapolation mode for elements outside the domain range.
            By default uses the method defined in fd. See extrapolation to
            more information.
        grid_points: Grid of points where
            the functions are evaluated in :func:`shift`.

    Returns:
        Functional data object with the registered samples.

    Examples:
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import (
        ...     landmark_shift_registration,
        ... )

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, random_state=1)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        >>> landmarks = landmarks.squeeze()

        The function will return the sample registered

        >>> landmark_shift_registration(fd, landmarks)
        FDataGrid(...)

    """
    shifts = landmark_shift_deltas(fd, landmarks, location=location)

    return fd.shift(
        shifts,
        restrict_domain=restrict_domain,
        extrapolation=extrapolation,
        grid_points=grid_points,
    )


def landmark_elastic_registration_warping(
    fd: FData,
    landmarks: ArrayLike,
    *,
    location: Optional[ArrayLike] = None,
    grid_points: Optional[GridPointsLike] = None,
) -> FDataGrid:
    """Calculate the transformation used in landmark registration.

        Let :math:`t_{ij}` the time where the sample :math:`i` has the feature
        :math:`j` and :math:`t^*_j` the new time for the feature. The warping
        function will transform the new time in the old time, i.e.,
        :math:`h_i(t^*_j)=t_{ij}`.
        The registered samples can be obtained as :math:`x^*_i(t)=x_i(h_i(t))`.

        See :footcite:`ramsay+silverman_2005_functional_landmark`
        for a detailed explanation.

    Args:
        fd: Functional data object.
        landmarks: List containing landmarks for each samples.
        location: Defines where
            the landmarks will be alligned. By default it will be used as
            location the mean of the landmarks.
        grid_points: Grid of points where
            the functions are evaluated to obtain a discrete
            representation of the object.

    Returns:
        FDataGrid with the warpings function needed to
        register the functional data object.

    Raises:
        ValueError: If the object to be registered has domain dimension greater
            than 1 or the list of landmarks or locations does not match with
            the number of samples.

    References:
        .. footbibliography::

    Examples:
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import (
        ...      landmark_elastic_registration_warping)

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, n_modes=2,
        ...                              random_state=9)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, n_modes=2,
        ...                                       random_state=9)
        >>> landmarks = landmarks.squeeze()

        The function will return the corresponding warping function

        >>> warping = landmark_elastic_registration_warping(fd, landmarks)
        >>> warping
        FDataGrid(...)

        The registered function can be obtained using function composition

        >>> fd.compose(warping)
        FDataGrid(...)

    """
    landmarks = np.asarray(landmarks)

    if fd.dim_domain > 1:
        raise NotImplementedError(
            "Method only implemented for objects with "
            "domain dimension up to 1.",
        )

    if len(landmarks) != fd.n_samples:
        raise ValueError(
            "The number of list of landmarks should be equal to "
            "the number of samples",
        )

    landmarks = landmarks.reshape((fd.n_samples, -1))

    location = (
        np.mean(landmarks, axis=0)
        if location is None
        else np.asarray(location)
    )

    assert isinstance(location, np.ndarray)

    n_landmarks = landmarks.shape[-1]

    data_matrix = np.empty((fd.n_samples, n_landmarks + 2))
    data_matrix[:, 0] = fd.domain_range[0][0]
    data_matrix[:, -1] = fd.domain_range[0][1]
    data_matrix[:, 1:-1] = landmarks

    if n_landmarks == len(location):
        if grid_points is None:
            grid_points = np.empty(n_landmarks + 2)
            grid_points[0] = fd.domain_range[0][0]
            grid_points[-1] = fd.domain_range[0][1]
            grid_points[1:-1] = location

    else:
        raise ValueError(
            f"Number of landmark locations should be equal than "
            f"the number of landmarks ({len(location)}) != ({n_landmarks})",
        )

    interpolation = SplineInterpolation(interpolation_order=3, monotone=True)

    warping = FDataGrid(
        data_matrix=data_matrix,
        grid_points=grid_points,
        interpolation=interpolation,
        extrapolation='bounds',
    )

    try:
        warping_points = fd.grid_points
    except AttributeError:
        warping_points = None

    return warping.to_grid(warping_points)


def landmark_registration(
    *args: Any,
    **kwargs: Any,
) -> FDataGrid:

    warnings.warn(
        "Function 'landmark_registration' has been renamed. "
        "Use 'landmark_elastic_registration' instead.",
        DeprecationWarning,
    )

    return landmark_elastic_registration(*args, **kwargs)


def landmark_elastic_registration(
    fd: FData,
    landmarks: ArrayLike,
    *,
    location: Optional[ArrayLike] = None,
    grid_points: Optional[GridPointsLike] = None,
) -> FDataGrid:
    """
    Perform landmark registration of the curves.

    Let :math:`t_{ij}` the time where the sample :math:`i` has the feature
    :math:`j` and :math:`t^*_j` the new time for the feature.
    The registered samples will have their features aligned, i.e.,
    :math:`x^*_i(t^*_j)=x_i(t_{ij})`.

    See :footcite:`ramsay+silverman_2005_functional_landmark`
    for a detailed explanation.

    Args:
        fd: Functional data object.
        landmarks: List containing landmarks for each samples.
        location: Defines where
            the landmarks will be alligned. By default it will be used as
            location the mean of the landmarks.
        grid_points: Grid of points where
            the functions are evaluated to obtain a discrete
            representation of the object. In case of objects with
            multidimensional :term:`domain` a list axis with points of
            evaluation for each dimension.

    Returns:
        FDataGrid with the functional data object registered.

    References:
        .. footbibliography::

    Examples:
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import (
        ...     landmark_elastic_registration,
        ... )
        >>> from skfda.representation.basis import BSpline

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, n_modes=2,
        ...                              random_state=9)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, n_modes=2,
        ...                                       random_state=9)
        >>> landmarks = landmarks.squeeze()

        The function will return the registered curves

        >>> landmark_elastic_registration(fd, landmarks)
        FDataGrid(...)

        This method will work for FDataBasis as for FDataGrids

        >>> fd = fd.to_basis(BSpline(n_basis=12))
        >>> landmark_elastic_registration(fd, landmarks)
        FDataGrid(...)

    """
    warping = landmark_elastic_registration_warping(
        fd,
        landmarks,
        location=location,
        grid_points=grid_points,
    )

    return fd.to_grid(grid_points).compose(warping)



from __future__ import annotations

from typing import Optional, TypeVar, Union

import numpy as np
from typing_extensions import Final

from ...representation import FData
from ...representation._typing import NDArrayFloat
from .._math import cosine_similarity, cosine_similarity_matrix
from ._utils import pairwise_metric_optimization

T = TypeVar("T", bound=Union[NDArrayFloat, FData])


class AngularDistance():
    r"""
    Calculate the angular distance between two objects.

    For each pair of observations x and y the angular distance between them is
    defined as the normalized "angle" between them:

    .. math::
        d(x, y) = \frac{\arccos \left(\frac{\langle x, y \rangle}{
        \sqrt{\langle x, x \rangle \langle y, y \rangle}} \right)}{\pi}

    where :math:`\langle {}\cdot{}, {}\cdot{} \rangle` is the inner product.
    This distance is defined in the interval [0, 1].

    Args:
        e1: First object.
        e2: Second object.

    Returns:
        Numpy vector where the i-th coordinate has the angular distance between
        the i-th element of the first object and the i-th element of the second
        one.

    Examples:
        Computes the angular distances between an object containing functional
        data corresponding to the functions y = 1 and y = x defined over the
        interval [0, 1] and another ones containing data of the functions y
        = 0 and y = x/2. The result then is an array of size 2 with the
        computed l2 distance between the functions in the same position in
        both.

        >>> import skfda
        >>> import numpy as np
        >>>
        >>> x = np.linspace(0, 1, 1001)
        >>> fd = skfda.FDataGrid([np.ones(len(x)), x], x)
        >>> fd2 =  skfda.FDataGrid([2*np.ones(len(x)), np.cos(x)], x)
        >>>
        >>> skfda.misc.metrics.angular_distance(fd, fd2).round(2)
        array([ 0.  ,  0.22])

    """

    def __call__(
        self,
        e1: T,
        e2: T,
    ) -> NDArrayFloat:
        """Compute the distance."""
        return np.arccos(cosine_similarity(e1, e2)) / np.pi

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}()"
        )


angular_distance: Final = AngularDistance()


@pairwise_metric_optimization.register
def _pairwise_metric_optimization_angular(
    metric: AngularDistance,
    elem1: Union[NDArrayFloat, FData],
    elem2: Optional[Union[NDArrayFloat, FData]],
) -> NDArrayFloat:

    return np.arccos(cosine_similarity_matrix(elem1, elem2)) / np.pi



from __future__ import annotations

from typing import Optional

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from .._typing import NDArrayFloat
from ._fdatabasis import FDataBasis


class CoefficientsTransformer(
    BaseEstimator,  # type:ignore
    TransformerMixin,  # type:ignore
):
    r"""
    Transformer returning the coefficients of FDataBasis objects as a matrix.

    Attributes:
        basis\_ (tuple): Basis used.

    Examples:
        >>> from skfda.representation.basis import (FDataBasis, Monomial,
        ...                                         CoefficientsTransformer)
        >>>
        >>> basis = Monomial(n_basis=4)
        >>> coefficients = [[0.5, 1, 2, .5], [1.5, 1, 4, .5]]
        >>> fd = FDataBasis(basis, coefficients)
        >>>
        >>> transformer = CoefficientsTransformer()
        >>> transformer.fit_transform(fd)
        array([[ 0.5,  1. ,  2. ,  0.5],
               [ 1.5,  1. ,  4. ,  0.5]])
                >>> transformer = CoefficientsTransformer(n_components=2)
        >>> transformer.fit_transform(fd)
        array([[ 0.5,  1. ],
               [ 1.5,  1. ]])

    """

    def __init__(self, n_components: Optional[int] = None) -> None:
        self.n_components = n_components

    def fit(  # noqa: D102
        self,
        X: FDataBasis,
        y: None = None,
    ) -> CoefficientsTransformer:

        self.basis_ = X.basis

        return self

    def transform(  # noqa: D102
        self,
        X: FDataBasis,
        y: None = None,
    ) -> NDArrayFloat:

        check_is_fitted(self)

        assert X.basis == self.basis_

        return X.coefficients[:, :self.n_components].copy()



"""Tests of the LinearDifferentialOperator."""

import unittest
from typing import Callable, Sequence, Union

import numpy as np

from skfda.misc.operators import LinearDifferentialOperator
from skfda.representation.basis import Constant, FDataBasis, Monomial

WeightCallable = Callable[[np.ndarray], np.ndarray]


class TestLinearDifferentialOperator(unittest.TestCase):
    """Tests of the linear differential operator."""

    def _assert_equal_weights(
        self,
        weights: Sequence[Union[float, WeightCallable]],
        weights2: Sequence[Union[float, WeightCallable]],
        msg: str,
    ) -> None:
        self.assertEqual(len(weights), len(weights2), msg)

        for w, w2 in zip(weights, weights2):

            eq = getattr(w, "equals", None)

            if eq is None:
                self.assertEqual(w, w2, msg)
            else:
                self.assertTrue(eq(w2), msg)

    def test_init_default(self) -> None:
        """Tests default initialization (do not penalize)."""
        lfd = LinearDifferentialOperator()
        weights = [0]

        self._assert_equal_weights(
            lfd.weights,
            weights,
            "Wrong list of weight functions of the linear operator",
        )

    def test_init_integer(self) -> None:
        """Tests initializations which only specify the order."""
        # Checks for a zero order Lfd object
        lfd_0 = LinearDifferentialOperator(order=0)
        weights = [1]

        self._assert_equal_weights(
            lfd_0.weights,
            weights,
            "Wrong list of weight functions of the linear operator",
        )

        # Checks for a non zero order Lfd object
        lfd_3 = LinearDifferentialOperator(3)
        weights = [0, 0, 0, 1]

        self._assert_equal_weights(
            lfd_3.weights,
            weights,
            "Wrong list of weight functions of the linear operator",
        )

        # Negative order must fail
        with np.testing.assert_raises(ValueError):
            LinearDifferentialOperator(-1)

    def test_init_list_int(self) -> None:
        """Tests initializations with integer weights."""
        weights = [1, 3, 4, 5, 6, 7]

        lfd = LinearDifferentialOperator(weights=weights)

        self._assert_equal_weights(
            lfd.weights,
            weights,
            "Wrong list of weight functions of the linear operator",
        )

    def test_init_list_fdatabasis(self) -> None:
        """Test initialization with functional weights."""
        n_basis = 4
        n_weights = 6

        monomial = Monomial(domain_range=(0, 1), n_basis=n_basis)

        weights = np.arange(n_basis * n_weights).reshape((n_weights, n_basis))

        fd = FDataBasis(monomial, weights)

        fdlist = [FDataBasis(monomial, w) for w in weights]
        lfd = LinearDifferentialOperator(weights=fdlist)

        self._assert_equal_weights(
            lfd.weights,
            list(fd),
            "Wrong list of weight functions of the linear operator",
        )

        # Check failure if intervals do not match
        constant = Constant(domain_range=(0, 2))
        fdlist.append(FDataBasis(constant, 1))
        with np.testing.assert_raises(ValueError):
            LinearDifferentialOperator(weights=fdlist)

    def test_init_wrong_params(self) -> None:
        """Check invalid parameters."""
        # Check specifying both arguments fail
        with np.testing.assert_raises(ValueError):
            LinearDifferentialOperator(1, weights=[1, 1])

        # Check invalid domain range
        monomial = Monomial(domain_range=(0, 1), n_basis=3)
        fdlist = [FDataBasis(monomial, [1, 2, 3])]

        with np.testing.assert_raises(ValueError):
            LinearDifferentialOperator(
                weights=fdlist,
                domain_range=(0, 2),
            )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""Linear smoother.

This module contains the abstract base class for all linear smoothers.

"""
from __future__ import annotations

import abc
from typing import Any, Mapping, Optional

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

from ... import FDataGrid
from ..._utils import _to_grid_points
from ...representation._typing import GridPointsLike


class _LinearSmoother(
    abc.ABC,
    BaseEstimator,  # type: ignore
    TransformerMixin,  # type: ignore
):
    """Linear smoother.

    Abstract base class for all linear smoothers. The subclasses must override
    ``hat_matrix`` to define the smoothing or 'hat' matrix.

    """

    def __init__(
        self,
        *,
        output_points: Optional[GridPointsLike] = None,
    ):
        self.output_points = output_points

    def hat_matrix(
        self,
        input_points: Optional[GridPointsLike] = None,
        output_points: Optional[GridPointsLike] = None,
    ) -> np.ndarray:

        # Use the fitted points if they are not provided
        if input_points is None:
            input_points = self.input_points_
        if output_points is None:
            output_points = self.output_points_

        return self._hat_matrix(
            input_points=self.input_points_,
            output_points=self.output_points_,
        )

    @abc.abstractmethod
    def _hat_matrix(
        self,
        input_points: GridPointsLike,
        output_points: GridPointsLike,
    ) -> np.ndarray:
        pass

    def _more_tags(self) -> Mapping[str, Any]:
        return {
            'X_types': [],
        }

    def fit(
        self,
        X: FDataGrid,
        y: None = None,
    ) -> _LinearSmoother:
        """Compute the hat matrix for the desired output points.

        Args:
            X: The data whose points are used to compute the matrix.
            y: Ignored.

        Returns:
            self

        """
        self.input_points_ = X.grid_points
        self.output_points_ = (
            _to_grid_points(self.output_points)
            if self.output_points is not None
            else self.input_points_
        )

        self.hat_matrix_ = self.hat_matrix()

        return self

    def transform(
        self,
        X: FDataGrid,
        y: None = None,
    ) -> FDataGrid:
        """Multiply the hat matrix with the function values to smooth them.

        Args:
            X: The data to smooth.
            y: Ignored

        Returns:
            Functional data smoothed.

        """
        assert all(
            np.array_equal(i, s) for i, s in zip(
                self.input_points_,
                X.grid_points,
            )
        )

        # The matrix is cached
        return X.copy(
            data_matrix=self.hat_matrix_ @ X.data_matrix,
            grid_points=self.output_points_,
        )

    def score(
        self,
        X: FDataGrid,
        y: FDataGrid,
    ) -> float:
        """Return the generalized cross validation (GCV) score.

        Args:
            X: The data to smooth.
            y: The target data. Typically the same as ``X``.

        Returns:
            Generalized cross validation score.

        """
        from .validation import LinearSmootherGeneralizedCVScorer

        return LinearSmootherGeneralizedCVScorer()(self, X, y)



import unittest

import numpy as np

import skfda


class TestsSklearn(unittest.TestCase):

    def setUp(self) -> None:
        unittest.TestCase.setUp(self)

        self.x = np.linspace(-1, 1, 1000)[:, np.newaxis]

    def _test_compare_sklearn(
        self,
        cov: skfda.misc.covariances.Covariance,
    ) -> None:
        cov_sklearn = cov.to_sklearn()
        cov_matrix = cov(self.x, self.x)
        cov_sklearn_matrix = cov_sklearn(self.x)

        np.testing.assert_array_almost_equal(cov_matrix, cov_sklearn_matrix)

    def test_linear(self) -> None:

        for variance in (1, 2):
            for intercept in (0, 1, 2):
                with self.subTest(variance=variance, intercept=intercept):
                    cov = skfda.misc.covariances.Linear(
                        variance=variance, intercept=intercept)
                    self._test_compare_sklearn(cov)

    def test_polynomial(self) -> None:

        for variance in (1, 2):
            for intercept in (0, 1, 2):
                for slope in (1, 2):
                    for degree in (1, 2, 3):
                        with self.subTest(
                            variance=variance,
                            intercept=intercept,
                            slope=slope,
                            degree=degree,
                        ):
                            cov = skfda.misc.covariances.Polynomial(
                                variance=variance,
                                intercept=intercept,
                                slope=slope,
                                degree=degree,
                            )
                            self._test_compare_sklearn(cov)

    def test_gaussian(self) -> None:

        for variance in (1, 2):
            for length_scale in (0.5, 1, 2):
                with self.subTest(
                    variance=variance,
                    length_scale=length_scale,
                ):
                    cov = skfda.misc.covariances.Gaussian(
                        variance=variance,
                        length_scale=length_scale,
                    )
                    self._test_compare_sklearn(cov)

    def test_exponential(self) -> None:

        for variance in (1, 2):
            for length_scale in (0.5, 1, 2):
                with self.subTest(
                    variance=variance,
                    length_scale=length_scale,
                ):
                    cov = skfda.misc.covariances.Exponential(
                        variance=variance,
                        length_scale=length_scale,
                    )
                    self._test_compare_sklearn(cov)

    def test_matern(self) -> None:

        for variance in (1, 2):
            for length_scale in (0.5, 1, 2):
                for nu in (0.5, 1, 1.5, 2, 2.5, 3.5, 4.5, np.inf):
                    with self.subTest(
                        variance=variance,
                        length_scale=length_scale,
                        nu=nu,
                    ):
                        cov = skfda.misc.covariances.Matern(
                            variance=variance,
                            length_scale=length_scale,
                            nu=nu,
                        )
                        self._test_compare_sklearn(cov)

    def test_white_noise(self) -> None:

        for variance in (1, 2):
            with self.subTest(variance=variance):
                cov = skfda.misc.covariances.WhiteNoise(variance=variance)
                self._test_compare_sklearn(cov)



"""
Scikit-fda and scikit-learn
===========================

In this section, we will explain how scikit-fda interacts  with the popular
machine learning package scikit-learn. We will introduce briefly the main
concepts of scikit-learn and how scikit-fda reuses the same concepts extending
them to the :term:`functional data analysis` field.

.. Disable isort
    isort:skip_file

"""

# Author: Carlos Ramos Carreño
# License: MIT

##############################################################################
# A brief summary of scikit-learn architecture
# --------------------------------------------
#
# The library `scikit-learn <https://scikit-learn.org/>`_ is probably the most
# well-known Python package for machine learning. This package focuses in
# machine learning using multivariate data, which should be stored in a numpy
# :class:`~numpy.ndarray` in order to process it. However, this library has
# defined a particular architecture that can be followed in order to provide
# new tools that work in situations not even imagined by the original authors,
# while remaining compatible with the tools already provided in scikit-learn.
#
# In scikit-fda, the same architecture is applied in order to work with
# functional data observations. As a result, scikit-fda tools are
# largely compatible with scikit-learn tools, and it is possible to reuse
# objects such as :class:`pipelines <sklearn.pipeline.Pipeline>` or even
# hyperparameter selection methods such as
# :class:`grid search cross-validation <sklearn.model_selection.GridSearchCV>`
# in the functional data setting.
#
# We will introduce briefly the main concepts in scikit-learn, and explain how
# the tools in scikit-fda are related with them. This is not intended as a full
# explanation of scikit-learn architecture, and the reader is encouraged to
# look at the `scikit-learn tutorials
# <https://scikit-learn.org/stable/tutorial/index.html>`_ in order to achieve
# a deeper understanding of it.

##############################################################################
# The Estimator object
# ^^^^^^^^^^^^^^^^^^^^
#
# A central concept in scikit-learn (and scikit-fda) is what is called an
# estimator. An estimator in this context is an object that can learn from
# the data. Thus, classification, regression and clustering methods, as well
# as transformations with parameters learned from the training data are
# particular kinds of estimators. Estimators can also be instanced passing
# parameters, which can be tuned to the data using hyperparameter selection
# methods.
#
# Estimator objects have a ``fit`` method, with receive the training data
# and (if necessary) the training targets. This method uses the training data
# in order to learn some parameters of a model. When the learned parameters
# are part of the user-facing API, then by convention they are attributes of
# the estimator ending in with the ``_`` character.

##############################################################################
# As a concrete example of this, consider a nearest centroid classifier
# for functional data. The object
# :class:`~skfda.ml.classification.NearestCentroid` is a classifier, and
# thus an estimator. As part of the training process the centroids of
# the classes are computed and available as the learned parameter
# ``centroids_``.
#
# .. note::
#     The function :func:`~sklearn.model_selection.train_test_split` is
#     one of the functions originally from scikit-learn that can be
#     directly reused in scikit-fda.

import skfda
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt

X, y = skfda.datasets.fetch_growth(return_X_y=True)

X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

classifier = skfda.ml.classification.NearestCentroid()
classifier.fit(X_train, y_train)
classifier.centroids_.plot()
plt.show()

##############################################################################
# Transformers
# ^^^^^^^^^^^^
#
# :term:`Transformers <sklearn:transformers>` are estimators which can convert
# data to a new form. Examples of them are preprocessing methods, such as
# smoothing, registration and dimensionality reduction methods. They always
# implement ``fit_transform`` for fitting and transforming the data in one
# step. The transformers may be :term:`sklearn:inductive`, which means that
# can transform new data using the learned parameters. In that case they
# implement the ``transform`` method to transform new data. If the
# transformation is reversible, they usually also implement
# ``ìnverse_transform``.

##############################################################################
# As an example consider the smoothing method
# :class:`skfda.preprocessing.smoothing.NadarayaWatson`. Smoothing methods
# attempt to remove noise from the data leveraging its continuous nature.
# As these methods discard information of the original data they usually are
# not reversible.

import skfda.preprocessing.smoothing.kernel_smoothers as ks

X, y = skfda.datasets.fetch_phoneme(return_X_y=True)

# Keep the first 5 functions
X = X[:5]

X.plot()

smoother = ks.NadarayaWatsonSmoother()
X_smooth = smoother.fit_transform(X)

X_smooth.plot()
plt.show()

##############################################################################
# Predictors (classifiers, regressors, clusterers...)
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# :term:`Predictors <sklearn:predictor>` in scikit-learn are estimators that
# can assign a certain target to a particular observation. This includes
# supervised methods such as classifiers (for which the target will be a class
# label), or regressors (for which the target is a real value, a vector, or,
# in functional data analysis, even a function!) and also unsupervised methods
# such as clusterers or outlying detector methods.
#
# Predictors should implement the ``fit_predict`` method for fitting the
# estimators and predicting the targets in one step and/or the ``predict``
# method for predicting the targets of possibly non previously bserved data.
# Usually :term:`sklearn:transductive` estimators implement only the former
# one, while :term:`sklearn:inductive` estimators implement the latter one (or
# both).
#
# Predictors can have additional non-mandatory methods, such as
# ``predict-proba`` for obtaining the probability of a particular prediction
# or ``score`` for evaluating the results of the prediction.

##############################################################################
# As an example, we can look at the :class:`~skfda.ml.clustering.KMeans`
# clustering method for functional data. This method will try to separate
# the data into different clusters according to the distance between
# observations.

X, y = skfda.datasets.fetch_weather(return_X_y=True)

# Use only the first value (temperature)
X = X.coordinates[0]

clusterer = skfda.ml.clustering.KMeans(n_clusters=3)
y_pred = clusterer.fit_predict(X)

X.plot(group=y_pred)
plt.show()

##############################################################################
# Metaestimators
# ^^^^^^^^^^^^^^
#
# In scikit-learn jargon, a :term:`sklearn:metaestimator` is an estimator
# that takes other estimators as parameters. There are several reasons for
# doing that, which will be explained now.

##############################################################################
# Composition metaestimators
# ++++++++++++++++++++++++++
#
# It is very common in machine learning to apply one or more preprocessing
# steps one after the other, before applying a final predictor. For this
# purpose scikit-learn offers the :class:`~sklearn.pipeline.Pipeline`, which
# join the steps together and uses the same estimator API for performing all
# steps in order (this is usually referred as the composite pattern in
# software engineering). The :class:`~sklearn.pipeline.Pipeline` estimator
# can be used with the functional data estimators available in scikit-fda.
# Moreover, as transformers such as dimensionality reduction methods can
# convert functional data to multivariate data usable by scikit-learn methods
# it is possible to mix methods from scikit-fda and scikit-learn in the same
# pipeline.
#
# .. warning::
#     In addition, scikit-learn offers estimators that can join several
#     transformations as new features of the same dataset (
#     :class:`~sklearn.pipeline.FeatureUnion`) or that can apply different
#     transformers to different columns of the data
#     (:class:`~sklearn.compose.ColumnTransformer`). These transformers
#     are not yet usable with functional data.

##############################################################################
# As an example, we can construct a pipeline that registers the data using
# shift registation, then applies a variable selection method to
# transform each observation to a 3D vector and then uses a SVM classifier
# to classify the data.

from skfda.preprocessing.dim_reduction import variable_selection as vs
from skfda.preprocessing.registration import LeastSquaresShiftRegistration
from sklearn.pipeline import Pipeline
from sklearn.svm import SVC

X, y = skfda.datasets.fetch_growth(return_X_y=True)

X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

pipeline = Pipeline([
    ("registration", LeastSquaresShiftRegistration()),
    ("dim_reduction", vs.RKHSVariableSelection(n_features_to_select=3)),
    ("classifier", SVC()),
])

pipeline.fit(X_train, y_train)
pipeline.score(X_test, y_test)

##############################################################################
# Hyperparameter optimizers
# +++++++++++++++++++++++++
#
# Some of the parameters used for the creation of an estimator need to be
# tuned to each particular dataset in order to improve the prediction accuracy
# and generalization. There are several techniques to do that already
# available in scikit-learn, such as grid search cross-validation
# (:class:`~sklearn.model_selection.GridSearchCV`) or randomized search
# (:class:`~sklearn.model_selection.RandomizedSearchCV`). As these
# hyperparameter optimizers only need to split the data and call ``score`` in
# the predictor, they can be directly used with the methods in scikit-fda.
#
# .. note::
#     In addition one could use any optimizer that understand the scikit-learn
#     API such as those in `scikit-optimize
#     <https://scikit-optimize.github.io>`_.

##############################################################################
# As an example, we will use :class:`~sklearn.model_selection.GridSearchCV`
# to select the number of neighbors used in a
# :class:`~skfda.ml.classification.KNeighborsClassifier`.

from sklearn.model_selection import GridSearchCV

X, y = skfda.datasets.fetch_growth(return_X_y=True)

X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

classifier = skfda.ml.classification.KNeighborsClassifier()

grid_search = GridSearchCV(
    estimator=classifier,
    param_grid={"n_neighbors": range(1, 10, 2)},
)

grid_search.fit(X_train, y_train)
n_neighbors = grid_search.best_estimator_.n_neighbors
score = grid_search.score(X_test, y_test)

print(n_neighbors, score)

##############################################################################
# Ensemble methods
# ++++++++++++++++
#
# The ensemble methods :class:`~sklearn.ensemble.VotingClassifier` and
# :class:`~sklearn.ensemble.VotingRegressor` in scikit-learn use several
# different estimators in order to predict the targets. As this is done
# by evaluating the passed estimators as black boxes, these predictors can
# also be combined with scikit-fda predictors.
#
# .. warning::
#     Other ensemble methods, such as
#     :class:`~sklearn.ensemble.BaggingClassifier` or
#     :class:`~sklearn.ensemble.AdaBoostClassifier` cannot yet
#     be used with functional data unless it has been
#     transformed to a multivariate dataset.

##############################################################################
# As an example we will use a voting classifier to classify data using as
# classifiers a knn-classifier, a nearest centroid classifier and a
# maximum depth classifier.

from sklearn.ensemble import VotingClassifier

X, y = skfda.datasets.fetch_growth(return_X_y=True)

X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

knn = skfda.ml.classification.KNeighborsClassifier()
nearest_centroid = skfda.ml.classification.NearestCentroid()
mdc = skfda.ml.classification.MaximumDepthClassifier()

voting = VotingClassifier([
    ("knn", knn),
    ("nearest_centroid", nearest_centroid),
    ("mdc", mdc),
])

voting.fit(X_train, y_train)
voting.score(X_test, y_test)

##############################################################################
# Multiclass and multioutput classification utilities
# +++++++++++++++++++++++++++++++++++++++++++++++++++
#
# The scikit-learn library also offers additional utilities that can convert
# a binary classifier into a multiclass classifier (such as
# :class:`~sklearn.multiclass.OneVsRestClassifier`) or to extend a single
# output classifier or regressor to accept also multioutput (vector-valued)
# targets.

##############################################################################
# In this example we want to use as a classifier the combination of a
# dimensionality reduction method (
# :class:`~skfda.preprocessing.dim_reduction.variable_selection.RKHSVariableSelection`)
# and a SVM classifier (:class:`~sklearn.svm.SVC`). As that particular
# dimensionality reduction method is only suitable for binary data, we use
# :class:`~sklearn.multiclass.OneVsRestClassifier` to classify in a
# multiclass dataset.

from sklearn.multiclass import OneVsRestClassifier

X, y = skfda.datasets.fetch_phoneme(return_X_y=True)

X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

pipeline = Pipeline([
    ("dim_reduction", vs.RKHSVariableSelection(n_features_to_select=3)),
    ("classifier", SVC()),
])

multiclass = OneVsRestClassifier(pipeline)

multiclass.fit(X_train, y_train)
multiclass.score(X_test, y_test)

##############################################################################
# Other scikit-learn utilities
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# In addition to the aforementioned objects, there are plenty of objects in
# scikit-learn that can be applied directly to functional data. We have
# already seen in the examples the function
# :func:`~sklearn.model_selection.train_test_split`. Other objects and
# functions such as :class:`~sklearn.model_selection.KFold` can be directly
# applied to functional data in order to split it into folds. Scorers for
# classification or regression, such as
# :func:`~sklearn.metrics.accuracy_score` can be directly applied to
# functional data problems.
#
# Moreover, there are plenty of libraries that aim to extend scikit-learn in
# several directions (take a look at the `list of related projects
# <https://scikit-learn.org/stable/related_projects.html>`_). You will
# probably see that a lot of the functionality can be applied to scikit-fda,
# as it uses the same API as scikit-learn.



from __future__ import annotations

from typing import Optional

import numpy as np
import scipy.integrate
from sklearn.base import BaseEstimator, TransformerMixin

from ..._utils import check_is_univariate
from ...representation import FDataGrid
from ...representation._typing import ArrayLike
from ._operators import Operator


class SRSF(
    Operator[FDataGrid, FDataGrid],
    BaseEstimator,  # type: ignore
    TransformerMixin,  # type: ignore
):
    r"""Square-Root Slope Function (SRSF) transform.

    Let :math:`f : [a,b] \rightarrow \mathbb{R}` be an absolutely continuous
    function, the SRSF transform is defined as

    .. math::
        SRSF(f(t)) = sgn(f(t)) \sqrt{|\dot f(t)|} = q(t)

    This representation it is used to compute the extended non-parametric
    Fisher-Rao distance between functions, wich under the SRSF representation
    becomes the usual :math:`\mathbb{L}^2` distance between functions.
    See :footcite:`srivastava+klassen_2016_analysis_square`.

    The inverse SRSF transform is defined as

    .. math::
        f(t) = f(a) + \int_{a}^t q(t)|q(t)|dt .

    This transformation is a mapping up to constant. Given the SRSF and the
    initial value :math:`f(a)` the original function can be obtained, for this
    reason it is necessary to store the value :math:`f(a)` during the fit,
    which is dropped due to derivation. If it is applied the inverse
    transformation without fit the estimator it is assumed that :math:`f(a)=0`.

    Args:
        eval_points: (array_like, optional): Set of points where the
            functions are evaluated, by default uses the sample points of
            the :class:`FDataGrid <skfda.FDataGrid>` transformed.
        initial_value (float, optional): Initial value to apply in the
            inverse transformation. If `None` there are stored the initial
            values of the functions during the transformation to apply
            during the inverse transformation. Defaults None.

    Attributes:
        eval_points: Set of points where the
            functions are evaluated, by default uses the grid points of the
            fdatagrid.
        initial_value: Initial value to apply in the
            inverse transformation. If `None` there are stored the initial
            values of the functions during the transformation to apply
            during the inverse transformation. Defaults None.

    Note:
        Due to the use of derivatives it is recommended that the samples are
        sufficiently smooth, or have passed a smoothing preprocessing before,
        in order to achieve good results.

    References:
        .. footbibliography::

    Examples:
        Create a toy dataset and apply the transformation and its inverse.

        >>> from skfda.datasets import make_sinusoidal_process
        >>> from skfda.misc.operators import SRSF
        >>> fd = make_sinusoidal_process(error_std=0, random_state=0)
        >>> srsf = SRSF()
        >>> srsf
        SRSF(...)

        Fits the estimator (to apply the inverse transform) and apply the SRSF

        >>> q = srsf.fit_transform(fd)

        Apply the inverse transform.

        >>> fd_pull_back = srsf.inverse_transform(q)

        The original and the pull back `fd` are almost equal

        >>> zero = fd - fd_pull_back
        >>> zero.data_matrix.flatten().round(3)
        array([ 0.,  0.,  0., ..., -0., -0., -0.])

    """

    def __init__(
        self,
        output_points: Optional[ArrayLike] = None,
        initial_value: Optional[float] = None,
    ) -> None:
        self.output_points = output_points
        self.initial_value = initial_value

    def __call__(self, vector: FDataGrid) -> FDataGrid:
        return self.fit_transform(vector)

    def fit(self, X: FDataGrid, y: None = None) -> SRSF:
        """
        Return self. This transformer does not need to be fitted.

        Args:
            X: Present for API conventions.
            y: Present for API conventions.

        Returns:
            (Estimator): self

        """
        return self

    def transform(self, X: FDataGrid, y: None = None) -> FDataGrid:
        r"""
        Compute the square-root slope function (SRSF) transform.

        Let :math:`f : [a,b] \rightarrow \mathbb{R}` be an absolutely
        continuous function, the SRSF transform is defined as
        :footcite:`srivastava+klassen_2016_analysis_square`:

        .. math::

            SRSF(f(t)) = sgn(f(t)) \sqrt{\dot f(t)|} = q(t)

        Args:
            X: Functions to be transformed.
            y: Present for API conventions.

        Returns:
            SRSF functions.

        Raises:
            ValueError: If functions are not univariate.

        """
        check_is_univariate(X)

        if self.output_points is None:
            output_points = X.grid_points[0]
        else:
            output_points = np.asarray(self.output_points)

        g = X.derivative()

        # Evaluation with the corresponding interpolation
        data_matrix = g(output_points)[..., 0]

        # SRSF(f) = sign(f) * sqrt|Df| (avoiding multiple allocation)
        sign_g = np.sign(data_matrix)
        data_matrix = np.abs(data_matrix, out=data_matrix)
        data_matrix = np.sqrt(data_matrix, out=data_matrix)
        data_matrix *= sign_g

        # Store the values of the transformation
        if self.initial_value is None:
            a = X.domain_range[0][0]
            self.initial_value_ = X(a).reshape(X.n_samples, 1, X.dim_codomain)

        return X.copy(data_matrix=data_matrix, grid_points=output_points)

    def inverse_transform(self, X: FDataGrid, y: None = None) -> FDataGrid:
        r"""
        Compute the inverse SRSF transform.

        Given the srsf and the initial value the original function can be
        obtained as :footcite:`srivastava+klassen_2016_analysis_square`:

        .. math::
            f(t) = f(a) + \int_{a}^t q(t)|q(t)|dt

        where :math:`q(t)=SRSF(f(t))`.

        If it is applied this inverse transformation without fitting the
        estimator it is assumed that :math:`f(a)=0`.

        Args:
            X: SRSF to be transformed.
            y: Present for API conventions.

        Returns:
            Functions in the original space.

        Raises:
            ValueError: If functions are multidimensional.
        """
        check_is_univariate(X)

        stored_initial_value = getattr(self, 'initial_value_', None)

        if self.initial_value is None and stored_initial_value is None:
            raise AttributeError(
                "When initial_value=None is expected a "
                "previous transformation of the data to "
                "store the initial values to apply in the "
                "inverse transformation. Also it is possible "
                "to fix these values setting the attribute"
                "initial value without a previous "
                "transformation.",
            )

        if self.output_points is None:
            output_points = X.grid_points[0]
        else:
            output_points = np.asarray(self.output_points)

        data_matrix = X(output_points)

        data_matrix *= np.abs(data_matrix)

        f_data_matrix = scipy.integrate.cumtrapz(
            data_matrix,
            x=output_points,
            axis=1,
            initial=0,
        )

        # If the transformer was fitted, sum the initial value
        if self.initial_value is None:
            f_data_matrix += self.initial_value_
        else:
            f_data_matrix += self.initial_value

        return X.copy(data_matrix=f_data_matrix, grid_points=output_points)



import unittest

import numpy as np

from skfda.ml.clustering import FuzzyCMeans, KMeans
from skfda.representation.grid import FDataGrid


class TestClustering(unittest.TestCase):

    # def setUp(self): could be defined for set up before any test

    def test_kmeans_univariate(self) -> None:
        data_matrix = [
            [1, 1, 2, 3, 2.5, 2],
            [0.5, 0.5, 1, 2, 1.5, 1],
            [-1, -1, -0.5, 1, 1, 0.5],
            [-0.5, -0.5, -0.5, -1, -1, -1],
        ]
        grid_points = [0, 2, 4, 6, 8, 10]
        fd = FDataGrid(data_matrix, grid_points)
        init = np.array([[0, 0, 0, 0, 0, 0], [2, 1, -1, 0.5, 0, -0.5]])
        init_fd = FDataGrid(init, grid_points)
        kmeans = KMeans(init=init_fd)
        distances_to_centers = kmeans.fit_transform(fd)
        np.testing.assert_allclose(
            distances_to_centers,
            np.array([
                [2.98142397, 9.23534876],
                [0.68718427, 6.50960828],
                [3.31243449, 4.39222798],
                [6.49679408, 0.0],
            ]),
        )
        np.testing.assert_array_equal(
            kmeans.predict(fd),
            np.array([0, 0, 0, 1]),
        )
        np.testing.assert_allclose(
            kmeans.transform(fd),
            np.array([[2.98142397, 9.23534876],
                      [0.68718427, 6.50960828],
                      [3.31243449, 4.39222798],
                      [6.49679408, 0.]]),
        )
        centers = FDataGrid(
            data_matrix=np.array([
                [0.16666667, 0.16666667, 0.83333333, 2., 1.66666667, 1.16666667],
                [-0.5, -0.5, -0.5, -1., -1., -1.],
            ]),
            grid_points=grid_points)
        np.testing.assert_array_almost_equal(
            kmeans.cluster_centers_.data_matrix,
            centers.data_matrix,
        )
        np.testing.assert_allclose(kmeans.score(fd), np.array([-20.33333333]))
        np.testing.assert_array_equal(kmeans.n_iter_, np.array([3.]))

    def test_fuzzy_kmeans_univariate(self) -> None:
        data_matrix = [
            [1, 1, 2, 3, 2.5, 2],
            [0.5, 0.5, 1, 2, 1.5, 1],
            [-1, -1, -0.5, 1, 1, 0.5],
            [-0.5, -0.5, -0.5, -1, -1, -1],
        ]
        grid_points = [0, 2, 4, 6, 8, 10]
        fd = FDataGrid(data_matrix, grid_points)
        fuzzy_kmeans = FuzzyCMeans()
        fuzzy_kmeans.fit(fd)
        np.testing.assert_array_equal(
            fuzzy_kmeans.predict_proba(fd).round(3),
            np.array([[0.965, 0.035],
                      [0.94, 0.06],
                      [0.227, 0.773],
                      [0.049, 0.951]]),
        )
        np.testing.assert_allclose(
            fuzzy_kmeans.transform(fd).round(3),
            np.array([[1.492, 7.879],
                      [1.294, 5.127],
                      [4.856, 2.633],
                      [7.775, 1.759]]),
        )
        centers = np.array([
            [0.707, 0.707, 1.455, 2.467, 1.981, 1.482],
            [-0.695, -0.695, -0.494, -0.197, -0.199, -0.398],
        ])
        np.testing.assert_allclose(
            fuzzy_kmeans.cluster_centers_.data_matrix[..., 0].round(3),
            centers)
        np.testing.assert_allclose(
            fuzzy_kmeans.score(fd),
            np.array([-12.025179]),
        )
        self.assertEqual(fuzzy_kmeans.n_iter_, 19)


if __name__ == '__main__':
    print()
    unittest.main()



from __future__ import annotations

import numpy as np
from sklearn.base import BaseEstimator, OutlierMixin

from ...representation import FDataGrid
from ..depth._depth import ModifiedBandDepth
from ..stats import modified_epigraph_index


class OutliergramOutlierDetector(
    BaseEstimator,  # type: ignore
    OutlierMixin,  # type: ignore
):
    r"""Outlier detector using the relation between MEI and MBD.

    Detects as outliers functions that have the vertical distance to the
    outliergram parabola greater than ``factor`` times the interquartile
    range (IQR) of those distances plus the third quartile. This corresponds
    to the points selected as outliers by the outliergram.

    Parameters:
        factor (float): The number of times the IQR is multiplied.

    Example:
        Function :math:`f : \mathbb{R}\longmapsto\mathbb{R}`.

        >>> import skfda
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 1, -1, 3, 2, 1],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 5, 5, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> data_matrix = [[0, 0, 0, 0, 0, 0],
        ...                [1, 1, 1, 1, 1, 1],
        ...                [2, 2, 2, 2, 2, 2],
        ...                [3, 3, 3, 3, 3, 3],
        ...                [9, 9, 9, -1, -1, -1],
        ...                [4, 4, 4, 4, 4, 4],
        ...                [5, 5, 5, 5, 5, 5],
        ...                [6, 6, 6, 6, 6, 6],
        ...                [7, 7, 7, 7, 7, 7],
        ...                [8, 8, 8, 8, 8, 8]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> out_detector = OutliergramOutlierDetector()
        >>> out_detector.fit_predict(fd)
        array([ 1,  1,  1,  1, -1,  1,  1,  1,  1,  1])

    """

    def __init__(self, *, factor: float = 1.5) -> None:
        self.factor = factor

    def _compute_parabola(self, X: FDataGrid) -> np.ndarray:
        """Compute the parabola in which pairs (mei, mbd) should lie."""
        a_0 = -2 / (X.n_samples * (X.n_samples - 1))
        a_1 = (2 * (X.n_samples + 1)) / (X.n_samples - 1)
        a_2 = a_0

        return (
            a_0 + a_1 * self.mei_
            + X.n_samples**2 * a_2 * self.mei_**2
        )

    def _compute_maximum_inlier_distance(self, distances: np.ndarray) -> float:
        """Compute the distance above which data are considered outliers."""
        first_quartile = np.percentile(distances, 25)  # noqa: WPS432
        third_quartile = np.percentile(distances, 75)  # noqa: WPS432
        iqr = third_quartile - first_quartile
        return third_quartile + self.factor * iqr

    def fit(self, X: FDataGrid, y: None = None) -> OutliergramOutlierDetector:
        self.mbd_ = ModifiedBandDepth()(X)
        self.mei_ = modified_epigraph_index(X)
        self.parabola_ = self._compute_parabola(X)
        self.distances_ = self.parabola_ - self.mbd_
        self.max_inlier_distance_ = self._compute_maximum_inlier_distance(
            self.distances_,
        )

        return self

    def fit_predict(self, X: FDataGrid, y: None = None) -> np.ndarray:
        self.fit(X, y)

        outliers = self.distances_ > self.max_inlier_distance_

        # Predict as scikit-learn outlier detectors
        predicted = ~outliers + outliers * -1

        return predicted



from __future__ import annotations

import abc
from functools import singledispatch
from typing import Any, Generic, TypeVar

import numpy as np

from ...misc._math import inner_product
from ...representation.basis import Basis, FDataBasis

CovariateType = TypeVar("CovariateType")


class CoefficientInfo(abc.ABC, Generic[CovariateType]):
    """
    Information about an estimated coefficient.

    Parameters:
        basis: Basis of the coefficient.

    """

    def __init__(
        self,
        basis: CovariateType,
    ) -> None:
        self.basis = basis

    @abc.abstractmethod
    def regression_matrix(
        self,
        X: CovariateType,
        y: np.ndarray,
    ) -> np.ndarray:
        """
        Return the constant coefficients matrix for regression.

        Parameters:
            X: covariate data for regression.
            y: target data for regression.

        Returns:
            Coefficients matrix.

        """
        pass

    @abc.abstractmethod
    def convert_from_constant_coefs(
        self,
        coefs: np.ndarray,
    ) -> CovariateType:
        """
        Return the coefficients object from the constant coefs.

        Parameters:
            coefs: estimated constant coefficients.

        Returns:
            Coefficient.

        """
        pass

    @abc.abstractmethod
    def inner_product(
        self,
        coefs: CovariateType,
        X: CovariateType,
    ) -> np.ndarray:
        """
        Inner product.

        Compute the inner product between the coefficient and
        the covariate.

        """
        pass


class CoefficientInfoNdarray(CoefficientInfo[np.ndarray]):

    def regression_matrix(  # noqa: D102
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> np.ndarray:

        return np.atleast_2d(X)

    def convert_from_constant_coefs(  # noqa: D102
        self,
        coefs: np.ndarray,
    ) -> np.ndarray:

        return coefs

    def inner_product(  # noqa: D102
        self,
        coefs: np.ndarray,
        X: np.ndarray,
    ) -> np.ndarray:

        return inner_product(coefs, X)


class CoefficientInfoFDataBasis(CoefficientInfo[FDataBasis]):
    """
    Information about a FDataBasis coefficient.

    Parameters:
        basis: Basis of the coefficient.

    """

    def regression_matrix(  # noqa: D102
        self,
        X: FDataBasis,
        y: np.ndarray,
    ) -> np.ndarray:
        # The matrix is the matrix of coefficients multiplied by
        # the matrix of inner products.

        xcoef = X.coefficients
        self.inner_basis = X.basis.inner_product_matrix(self.basis)
        return xcoef @ self.inner_basis

    def convert_from_constant_coefs(  # noqa: D102
        self,
        coefs: np.ndarray,
    ) -> FDataBasis:
        return FDataBasis(self.basis.basis, coefs.T)

    def inner_product(  # noqa: D102
        self,
        coefs: FDataBasis,
        X: FDataBasis,
    ) -> np.ndarray:
        # Efficient implementation of the inner product using the
        # inner product matrix previously computed
        return inner_product(coefs, X, inner_product_matrix=self.inner_basis.T)


@singledispatch
def coefficient_info_from_covariate(
    X: CovariateType,
    y: np.ndarray,
    **_: Any,
) -> CoefficientInfo[CovariateType]:
    """Make a coefficient info object from a covariate."""
    raise ValueError(f"Invalid type of covariate = {type(X)}.")


@coefficient_info_from_covariate.register(np.ndarray)
def _coefficient_info_from_covariate_ndarray(
    X: np.ndarray,
    y: np.ndarray,
    **_: Any,
) -> CoefficientInfo[np.ndarray]:
    return CoefficientInfoNdarray(basis=np.identity(X.shape[1], dtype=X.dtype))


@coefficient_info_from_covariate.register(FDataBasis)
def _coefficient_info_from_covariate_fdatabasis(
    X: FDataBasis,
    y: np.ndarray,
    *,
    basis: Basis,
    **_: Any,
) -> CoefficientInfoFDataBasis:

    if basis is None:
        basis = X.basis

    if not isinstance(basis, Basis):
        raise TypeError(f"basis must be a Basis object, not {type(basis)}")

    return CoefficientInfoFDataBasis(basis=basis.to_basis())



"""
Module to interpolate functional data objects.
"""
from __future__ import annotations

import abc
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Sequence,
    Tuple,
    Union,
    cast,
)

import numpy as np

from scipy.interpolate import (
    PchipInterpolator,
    RectBivariateSpline,
    RegularGridInterpolator,
    UnivariateSpline,
)

from .._utils import _to_array_maybe_ragged
from ._typing import ArrayLike
from .evaluator import Evaluator

if TYPE_CHECKING:
    from . import FData

SplineCallable = Callable[..., np.ndarray]


class _SplineList(abc.ABC):
    """ABC for list of interpolations."""

    def __init__(
        self,
        fdatagrid: FData,
        interpolation_order: Union[int, Sequence[int]] = 1,
        smoothness_parameter: float = 0,
    ):

        super().__init__()

        self.fdatagrid = fdatagrid
        self.interpolation_order = interpolation_order
        self.smoothness_parameter = smoothness_parameter
        self.splines: Sequence[Sequence[SplineCallable]]

    # @abc.abstractmethod
    # @property
    # def splines(self) -> Sequence[SplineCallable]:
        # pass

    @abc.abstractmethod
    def _evaluate_one(
        self,
        spline: SplineCallable,
        eval_points: np.ndarray,
    ) -> np.ndarray:
        """Evaluate one spline of the list."""
        pass

    def _evaluate_codomain(
        self,
        spline_list: Sequence[SplineCallable],
        eval_points: np.ndarray,
    ) -> np.ndarray:
        """Evaluate a multidimensional sample."""
        return np.array([
            self._evaluate_one(spl, eval_points)
            for spl in spline_list
        ]).T

    def evaluate(
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:

        res: np.ndarray

        if aligned:

            eval_points = np.asarray(eval_points)

            # Points evaluated inside the domain
            res = np.apply_along_axis(
                self._evaluate_codomain,
                1,
                self.splines,
                eval_points,
            )

            res = res.reshape(
                fdata.n_samples,
                eval_points.shape[0],
                fdata.dim_codomain,
            )

        else:
            eval_points = cast(Iterable[ArrayLike], eval_points)

            res = _to_array_maybe_ragged([
                self._evaluate_codomain(s, np.asarray(e))
                for s, e in zip(self.splines, eval_points)
            ])

        return res


class _SplineList1D(_SplineList):
    """List of interpolations for curves.

    List of interpolations for objects with domain
    dimension = 1. Calling internally during the creation of the
    evaluator.

    Uses internally the scipy interpolation UnivariateSpline or
    PchipInterpolator.

    Args:
        fdatagrid (FDatagrid): Fdatagrid to interpolate.
        interpolation_order (int, optional): Order of the interpolation, 1
            for linear interpolation, 2 for cuadratic, 3 for cubic and so
            on. In case of curves and surfaces there is available
            interpolation up to degree 5. For higher dimensional objects
            only linear or nearest interpolation is available. Default
            lineal interpolation.
        smoothness_parameter (float, optional): Penalisation to perform
            smoothness interpolation. Option only available for curves and
            surfaces. If 0 the residuals of the interpolation will be 0.
            Defaults 0.
        monotone (boolean, optional): Performs monotone interpolation in
            curves using a PCHIP interpolator. Only valid for curves (domain
            dimension equal to 1) and interpolation order equal to 1 or 3.
            Defaults false.

    Returns:
        (np.ndarray): Array of size n_samples x dim_codomain with the
        corresponding interpolation of the sample i, and image dimension j
        in the entry (i,j) of the array.

    Raises:
        ValueError: If the value of the interpolation k is not valid.

    """

    def __init__(
        self,
        fdatagrid: FData,
        interpolation_order: Union[int, Sequence[int]] = 1,
        smoothness_parameter: float = 0,
        monotone: bool = False,
    ):

        super().__init__(
            fdatagrid=fdatagrid,
            interpolation_order=interpolation_order,
            smoothness_parameter=smoothness_parameter,
        )

        self.monotone = monotone

        if (
            isinstance(self.interpolation_order, Sequence)
            or not 1 <= self.interpolation_order <= 5
        ):
            raise ValueError(
                f"Invalid degree of interpolation "
                f"({self.interpolation_order}). Must be "
                f"an integer greater than 0 and lower or "
                f"equal than 5.",
            )

        if self.monotone and self.smoothness_parameter != 0:
            raise ValueError(
                "Smoothing interpolation is not supported with "
                "monotone interpolation",
            )

        if self.monotone and self.interpolation_order in {2, 4}:
            raise ValueError(
                f"monotone interpolation of degree "
                f"{self.interpolation_order}"
                f"not supported.",
            )

        # Monotone interpolation of degree 1 is performed with linear spline
        monotone = self.monotone
        if self.monotone and self.interpolation_order == 1:
            monotone = False

        grid_points = fdatagrid.grid_points[0]

        if monotone:
            def constructor(  # noqa: WPS430
                data: np.ndarray,
            ) -> SplineCallable:
                """Construct an unidimensional cubic monotone interpolation."""
                return PchipInterpolator(grid_points, data)

        else:

            def constructor(  # noqa: WPS430, WPS440
                data: np.ndarray,
            ) -> SplineCallable:
                """Construct an unidimensional interpolation."""
                return UnivariateSpline(
                    grid_points,
                    data,
                    s=self.smoothness_parameter,
                    k=self.interpolation_order,
                )

        self.splines = np.apply_along_axis(
            constructor,
            1,
            fdatagrid.data_matrix,
        )

    def _evaluate_one(
        self,
        spline: SplineCallable,
        eval_points: np.ndarray,
    ) -> np.ndarray:
        try:
            return spline(eval_points)[:, 0]
        except ValueError:
            return np.zeros_like(eval_points)


class _SplineList2D(_SplineList):
    """List of interpolations for surfaces.

    List of interpolations for objects with domain
    dimension = 2. Calling internally during the creationg of the
    evaluator.

    Uses internally the scipy interpolation RectBivariateSpline.

    Args:
        fdatagrid (FDatagrid): Fdatagrid to interpolate.
        interpolation_order (int, optional): Order of the interpolation, 1
            for linear interpolation, 2 for cuadratic, 3 for cubic and so
            on. In case of curves and surfaces there is available
            interpolation up to degree 5. For higher dimensional objects
            only linear or nearest interpolation is available. Default
            lineal interpolation.
        smoothness_parameter (float, optional): Penalisation to perform
            smoothness interpolation. Option only available for curves and
            surfaces. If 0 the residuals of the interpolation will be 0.
            Defaults 0.
        monotone (boolean, optional): Performs monotone interpolation in
            curves using a PCHIP interpolator. Only valid for curves (domain
            dimension equal to 1) and interpolation order equal to 1 or 3.
            Defaults false.

    Returns:
        (np.ndarray): Array of size n_samples x dim_codomain with the
        corresponding interpolation of the sample i, and image dimension j
        in the entry (i,j) of the array.

    Raises:
        ValueError: If the value of the interpolation k is not valid.

    """

    def __init__(
        self,
        fdatagrid: FData,
        interpolation_order: Union[int, Sequence[int]] = 1,
        smoothness_parameter: float = 0,
    ):

        super().__init__(
            fdatagrid=fdatagrid,
            interpolation_order=interpolation_order,
            smoothness_parameter=smoothness_parameter,
        )

        if isinstance(self.interpolation_order, int):
            kx = self.interpolation_order
            ky = kx
        elif len(self.interpolation_order) == 2:
            kx = self.interpolation_order[0]
            ky = self.interpolation_order[1]
        else:
            raise ValueError("k should be numeric or a tuple of length 2.")

        if kx > 5 or kx <= 0 or ky > 5 or ky <= 0:
            raise ValueError(
                f"Invalid degree of interpolation ({kx},{ky}). "
                f"Must be an integer greater than 0 and lower or "
                f"equal than 5.",
            )

        # Matrix of splines
        splines = np.empty(
            (fdatagrid.n_samples, fdatagrid.dim_codomain),
            dtype=object,
        )

        for i in range(fdatagrid.n_samples):
            for j in range(fdatagrid.dim_codomain):
                splines[i, j] = RectBivariateSpline(
                    fdatagrid.grid_points[0],
                    fdatagrid.grid_points[1],
                    fdatagrid.data_matrix[i, :, :, j],
                    kx=kx,
                    ky=ky,
                    s=self.smoothness_parameter,
                )

        self.splines = splines

    def _evaluate_one(
        self,
        spline: SplineCallable,
        eval_points: np.ndarray,
    ) -> np.ndarray:

        return spline(
            eval_points[:, 0],
            eval_points[:, 1],
            grid=False,
        )


class _SplineListND(_SplineList):
    """
    List of interpolations.

    List of interpolations for objects with domain
    dimension > 2. Calling internally during the creationg of the
    evaluator.

    Only linear and nearest interpolations are available for objects with
    domain dimension >= 3. Uses internally the scipy interpolation
    RegularGridInterpolator.

    Args:
        grid_points (np.ndarray): Sample points of the fdatagrid.
        data_matrix (np.ndarray): Data matrix of the fdatagrid.
        k (integer): Order of the spline interpolations.

    Returns:
        (np.ndarray): Array of size n_samples x dim_codomain with the
        corresponding interpolation of the sample i, and image dimension j
        in the entry (i,j) of the array.

    Raises:
        ValueError: If the value of the interpolation k is not valid.

    """

    def __init__(
        self,
        fdatagrid: FData,
        interpolation_order: Union[int, Sequence[int]] = 1,
        smoothness_parameter: float = 0,
    ) -> None:
        super().__init__(
            fdatagrid=fdatagrid,
            interpolation_order=interpolation_order,
            smoothness_parameter=smoothness_parameter,
        )

        if self.smoothness_parameter != 0:
            raise ValueError(
                "Smoothing interpolation is only supported with "
                "domain dimension up to 2.",
            )

        # Parses method of interpolation
        if self.interpolation_order == 0:
            method = 'nearest'
        elif self.interpolation_order == 1:
            method = 'linear'
        else:
            raise ValueError(
                "interpolation order should be 0 (nearest) or 1 (linear).",
            )

        splines = np.empty(
            (fdatagrid.n_samples, fdatagrid.dim_codomain),
            dtype=object,
        )

        for i in range(fdatagrid.n_samples):
            for j in range(fdatagrid.dim_codomain):
                splines[i, j] = RegularGridInterpolator(
                    fdatagrid.grid_points,
                    fdatagrid.data_matrix[i, ..., j],
                    method=method,
                    bounds_error=False,
                )

        self.splines = splines

    def _evaluate_one(
        self,
        spline: SplineCallable,
        eval_points: np.ndarray,
    ) -> np.ndarray:

        return spline(eval_points)


class SplineInterpolation(Evaluator):
    """
    Spline interpolation.

    Spline interpolation of discretized functional objects. Implements
    different interpolation methods based in splines, using the sample
    points of the grid as nodes to interpolate.

    See the interpolation example to a detailled explanation.

    Attributes:
        interpolation_order (int, optional): Order of the interpolation, 1
            for linear interpolation, 2 for cuadratic, 3 for cubic and so
            on. In case of curves and surfaces there is available
            interpolation up to degree 5. For higher dimensional objects
            only linear or nearest interpolation is available. Default
            lineal interpolation.
        smoothness_parameter (float, optional): Penalisation to perform
            smoothness interpolation. Option only available for curves and
            surfaces. If 0 the residuals of the interpolation will be 0.
            Defaults 0.
        monotone (boolean, optional): Performs monotone interpolation in
            curves using a PCHIP interpolator. Only valid for curves (domain
            dimension equal to 1) and interpolation order equal to 1 or 3.
            Defaults false.

    """

    def __init__(
        self,
        interpolation_order: Union[int, Sequence[int]] = 1,
        *,
        smoothness_parameter: float = 0,
        monotone: bool = False,
    ) -> None:
        self._interpolation_order = interpolation_order
        self._smoothness_parameter = smoothness_parameter
        self._monotone = monotone

    @property
    def interpolation_order(self) -> Union[int, Tuple[int, ...]]:
        """Interpolation order."""

        return (
            self._interpolation_order
            if isinstance(self._interpolation_order, int)
            else tuple(self._interpolation_order)
        )

    @property
    def smoothness_parameter(self) -> float:
        """Smoothness parameter."""
        return self._smoothness_parameter

    @property
    def monotone(self) -> bool:
        """Flag to perform monotone interpolation."""
        return self._monotone

    def _build_interpolator(
        self,
        fdatagrid: FData,
    ) -> _SplineList:

        if fdatagrid.dim_domain == 1:
            return _SplineList1D(
                fdatagrid=fdatagrid,
                interpolation_order=self.interpolation_order,
                smoothness_parameter=self.smoothness_parameter,
                monotone=self.monotone,
            )

        elif self.monotone:
            raise ValueError(
                "Monotone interpolation is only supported with "
                "domain dimension equal to 1.",
            )

        elif fdatagrid.dim_domain == 2:
            return _SplineList2D(
                fdatagrid=fdatagrid,
                interpolation_order=self.interpolation_order,
                smoothness_parameter=self.smoothness_parameter,
            )

        return _SplineListND(
            fdatagrid=fdatagrid,
            interpolation_order=self.interpolation_order,
            smoothness_parameter=self.smoothness_parameter,
        )

    def _evaluate(  # noqa: D102
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:

        spline_list = self._build_interpolator(fdata)

        return spline_list.evaluate(fdata, eval_points, aligned=aligned)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"interpolation_order={self.interpolation_order}, "
            f"smoothness_parameter={self.smoothness_parameter}, "
            f"monotone={self.monotone})"
        )

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and self.interpolation_order == other.interpolation_order
            and self.smoothness_parameter == other.smoothness_parameter
            and self.monotone == other.monotone
        )
